/* ======================================================================
 *
 *  CCCCC          serialspy.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: January 5, 2000
 * CC   CC         Last Modified: March 3, 2000
 *  CCCCC
 *
 * Code file for spying on the bytes going through (ie. coming from)
 *   a serial cable.  Intended to help make sure serial devices are
 *   functioning properly, and also to help reverse engineer data
 *   protocols used by said devices.
 *
 * Copyright 2000, Bill Sherman & Friends, All rights reserved
 * With the intent to provide an open-source licence to be named later
 * ====================================================================== */
/*************************************************************************

USAGE:
	serialspy [-l <line>] [-b <baud>] [-o <output port>] [-f <frequency>]

	While running, the following keys can be used to adjust some parameters:
		'q' - quit
		'+' - increase the frequency
		'-' - decrease the frequency
		'>' - shift the phase to the right
		'<' - shift the phase to the left

HISTORY:
	5 January 2000 -- First version written by Bill Sherman, with help
		from Stuart Levy on the non-blocked reading of characters
		from the stdin terminal.  Tested on SGI IRIX 6.2 (works!),
		and Linux Redhat 6.0 (also works!).

TODO:

	- add the ability to toggle the printing of chars to the terminal
		(with a CLA to begin that way)

	- add the ability to read from one serial port, and write to another
		(of course, it's a two way deal, but you know what I mean).

*************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#if 0
#include <asm/ioctls.h>
#endif
#include <fcntl.h>

#if 1	/* this seems to be required for Linux (at least post RH 5.2 or 6.0) */
#define FIONREAD	0x541B
#endif

#define TEST_APP
#include "vr_serial.h"
#include "vr_debug.h"


/* We need to do this because the defines in vr_debug.h will */
/*   affect how the vr_serial.c code is compiled.            */
#include "vr_serial.c"

/* for some reason, this must come after vr_serial.c, or it gets undef'd */
#include <unistd.h>	/* for STDIN_FILENO */

char	*binary(unsigned int val, int len, char *str);

main(int argc, char *argv[])
{
	/* CLA parsing information */
static	char	*err_usage = "Usage: %s [-l <line>] [-b <baud>] [-f <frequency>]\n";
static	char	*err_unk_opt = "%s: unknown option '%s'.\n";
static	char	*err_missing_arg = "%s: missing argument to '%s'.\n";
static	char	*progname;

	/* the rest of the variables */
static	char	buffer[2048];
static	char	bin_buf[32];
#if defined(__linux)
	char	*inport = "/dev/ttyS0";		/* default serial port on Linux */
#else
	char	*inport = "/dev/ttyd2";		/* default serial port on SGI & other */
#endif
	char	*outport = NULL;
	int	fdin = -1;
	int	fdout = -1;
	int	baud_int = 9600;
	int	baud_enum = vrSerialBaudIntToEnum(baud_int);
	int	print_hex = 1;
	int	print_binary = 0;
	int	print_ascii = 1;
	int	frequency = 16;
	int	phase_shift = 0;
	int	bytes_read = 0;
	int	bytes_to_print;
	int	bytes_printed;		/* number of bytes printed since last read */
	int	need_separator = 0;
	int	count;
	char	byte;
unsigned char	input;
struct	termios	stdin_tios;
	int	inbytes_waiting;
	char	intr_char = '\3';
	char	store_vtime;
	char	store_vmin;
	int	done = 0;

	/******************/
	/* parse the CLAs */
	progname = argv[0];
	while ((argc > 1) && (argv[1][0] == '-')) {
		if (!strcmp(argv[1], "-l")) {		/** input port (line) **/
			argv++; argc--;

			if (argc !=1) {
				inport = strdup(argv[1]);
				fprintf(stderr, "Connecting to device on port '%s'.\n", inport);
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, err_usage, progname);
				exit(1);
			}

		} else if (!strcmp(argv[1], "-b")) {	/** baud rate **/
			argv++; argc--;

			if (argc !=1) {
				baud_int = atoi(argv[1]);
				baud_enum = vrSerialBaudIntToEnum(baud_int);
				fprintf(stderr, "Connecting to device at %d baud.\n", baud_int);
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, err_usage, progname);
				exit(1);
			}

		} else if (!strcmp(argv[1], "-f")) {	/** print frequency **/
			argv++; argc--;

			if (argc !=1) {
				frequency = atoi(argv[1]);
				fprintf(stderr, "Will begin printing with %d bytes per line.\n", frequency);
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, err_usage, progname);
				exit(1);
			}

		} else if (!strcmp(argv[1], "-help") || !strcmp(argv[1], "-usage")) {

			/* show usage and exit */
			fprintf(stderr, err_usage, progname);
			exit(2);

		} else {

			/* state unknown option */
			fprintf(stderr, err_unk_opt, progname, argv[1]);
			fprintf(stderr, err_usage, progname);
		}

		argv++; argc--;
	}

	if (argc < 1) {
		fprintf(stderr, err_usage, progname);
		exit(1);
	}


	/***************************************************************/
	/* open the "incoming" port (ie. port connected to the device) */
	fdin = vrSerialOpen(inport, baud_enum);
	if (fdin < 0) {
		fprintf(stderr, RED_TEXT "Unable to open serial port '%s'\n" NORM_TEXT, inport);
		exit(1);
	}

	/*********************************************************************************/
	/* setup standard in to be non-canonical, and thus we can get chars w/o blocking */

	/* get the current parameters */
	tcgetattr(STDIN_FILENO, &stdin_tios);

	/* turn off canonical processing and the echoing of characters */
	stdin_tios.c_lflag &= ~(ICANON | ECHO);

	/* setup for non-blocking mode */
	store_vmin = stdin_tios.c_cc[VMIN];
	store_vtime = stdin_tios.c_cc[VTIME];
	stdin_tios.c_cc[VMIN] = 0;
	stdin_tios.c_cc[VTIME] = 0;

	/* make the new parameters take affect */
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &stdin_tios);

	intr_char = stdin_tios.c_cc[VINTR];

#if 0
#define FIONREAD	0x541B
#endif
	/**************************/
	/* loop until interrupted */
	while(!done) {
		/******************************************************/
		/* read the terminal for any inputs, and process them */
		while(ioctl(STDIN_FILENO, FIONREAD, &inbytes_waiting) == 0 && inbytes_waiting > 0 && read(STDIN_FILENO, &input, 1) > 0) {
#if 0
			printf("input char = %02x\n", input);
#endif
			/* this doesn't seem to work, but no harm */
			if (input == intr_char)
				done = 1;

			switch (input) {
			case '\3':
			case 'q':
			case 'Q':
				done = 1;
				break;
			case '+':
			case '=':
				frequency++;
				break;
			case '-':
			case '_':
				frequency--;
				if (frequency <= 0)
					frequency = 1;
				break;
			case '>':
			case '.':
				phase_shift--;
				break;
			case '<':
			case ',':
				phase_shift++;
				break;

			case 'a':
			case 'A':
				print_ascii ^= 1;
				break;
			case 'b':
			case 'B':
				print_binary ^= 1;
				break;
			case 'h':
			case 'H':
				print_hex ^= 1;
				break;
			}
		}

		/******************************************/
		/* now read from the incoming serial port */
		bytes_read += (ssize_t)vrSerialRead(fdin, &buffer[bytes_read], sizeof(buffer)-bytes_read);
		bytes_printed = 0;
		bytes_to_print = frequency + phase_shift;

		while(bytes_read > bytes_printed + bytes_to_print) {
			need_separator = 0;
			phase_shift = 0;
#if 0
			printf(BOLD_TEXT "%d: ", bytes_printed);
#endif

			/************************/
			/* print the HEX values */
			if (print_hex) {
				printf(NORM_TEXT "%d: ", bytes_to_print);
				for (count = 0; count < bytes_to_print; count++) {
					byte = buffer[count+bytes_printed] & 0x7f;
					if (byte != buffer[count+bytes_printed] & 0x7f)
						printf(RED_TEXT);
					else	printf(NORM_TEXT);
					printf("%02x ", (unsigned int)(buffer[count+bytes_printed] & 0xff));
				}

				need_separator = 1;
			}

			/***************************/
			/* print the BINARY values */
			if (print_binary) {
				if (need_separator)
					printf(BOLD_TEXT " |  " NORM_TEXT);
				for (count = 0; count < bytes_to_print; count++) {
					byte = buffer[count+bytes_printed] & 0x7f;
					if (byte != buffer[count+bytes_printed] & 0x7f)
						printf(RED_TEXT);
					else	printf(NORM_TEXT);
					printf("%s ", binary((unsigned int)(buffer[count+bytes_printed] & 0xff), 8, bin_buf));
				}
				need_separator = 1;
			}

			/**************************/
			/* print the ASCII values */
			if (print_ascii) {
				if (need_separator)
					printf(BOLD_TEXT " |  " NORM_TEXT);
				for (count = 0; count < bytes_to_print; count++) {
					byte = buffer[count+bytes_printed] & 0x7f;
					if (byte != buffer[count+bytes_printed] & 0x7f)
						printf(RED_TEXT);
					else	printf(NORM_TEXT);
					printf("%c ", (isprint(byte) ? byte : '.'));
				}
				need_separator = 1;
			}

			printf(NORM_TEXT "\n");
			bytes_printed += count;
		}

		/* shift the non-printed stuff to beginning of buffer */
		bytes_read -= bytes_printed;
		memmove(buffer, buffer+bytes_printed, bytes_read);
	}

	/********************************/
	/* restore stdin as best we can */
	stdin_tios.c_lflag |= ICANON | ECHO;
	stdin_tios.c_cc[VMIN] = store_vmin;
	stdin_tios.c_cc[VTIME] = store_vtime;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &stdin_tios);
}


/*******************************************************************/
/* binary(): takes an integer, a number of bits, and a pointer to  */
/*   memory in which to store the result.  The resultant string    */
/*   is placed in the memory, and returned.  The string is made up */
/*   of the least significant N bits, based on the value of len.   */
/*******************************************************************/
char *binary(unsigned int val, int len, char *str)
{
	int     count;

	str[len] = '\0';
	for (count = 0; count < len; count++) {
		str[len-count-1] = (((1 << count) & val) ? '1' : '0');
	}
	return(str);
}



