/* ======================================================================
 * 
 *  CCCCC          valtest.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: March 14, 2000
 * CC   CC         Last Modified: September 9, 2002
 *  CCCCC
 * 
 * Code file for a sample VR application with multiple valuators.
 *  (initially created to test the cyberglove interface.)
 *  (started as a copy of travel.c)
 * 
 * Copyright 2002, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

#include "freevr.h"


/* these functions will be used as arguments */
void	init_gfx();
void	draw_valuator_world();
void	exit_application();
void	update_world();

#define PERSONAL_SIMULATOR
#ifdef PERSONAL_SIMULATOR
void	my_sim(vrRenderInfo *renderinfo, int mask);
#endif


int main(int argc, char* argv[])
{

	signal(SIGINT, SIG_IGN);

	vrConfigure(&argc, argv, NULL);
	vrStart();

	signal(SIGINT, exit_application);

	/* TODO: it would be nice to put these above vrStart(), since */
	/*   it will block until all input devices are open, and this */
	/*   way we could be rendering something while waiting.       */
	/* TODO: actually, we almost *need* to have the init_gfx stuff*/
	/*   set before vrStart, so the window opening function will  */
	/*   have something to call (other than vrDoNothing()).       */
	/*   In fact, I have no idea how init_gfx() gets called at    */
	/*   all given the current setup.                             */
	vrSetFunc(VRFUNC_ALL_DISPLAY_INIT, vrCreateCallback(init_gfx, 0));
	vrSetFunc(VRFUNC_ALL_DISPLAY, vrCreateCallback(draw_valuator_world, 0));
#ifdef PERSONAL_SIMULATOR
	vrSetFunc(VRFUNC_ALL_DISPLAY_SIM, vrCreateCallback(my_sim, 0));
#endif
	vrSetFunc(VRFUNC_HANDLE_USR2, vrCreateCallback(vrDbgInfo, 0));

	if (vrContext->input->num_input_devices > 0) {
		vrDbgPrintfN(SELDOM_DBGLVL+1, "VALTEST: input device[0] version = '%s'\n",
			vrContext->input->input_devices[0]->version);
	} else {
		vrDbgPrintfN(SELDOM_DBGLVL+1, "VALTEST: No input devices defined.\n");
	}

	if (vrDbgDo(AALWAYS_DBGLVL+1)) {
		vrFprintContext(stdout, vrContext, verbose);
		vrFprintConfig(stdout, vrContext->config, verbose);
		vrFprintInput(stdout, vrContext->input, verbose);
	}

	vrDbgPrintf("VALTEST: input device[0] version = '%s'\n", 
		vrContext->input->input_devices[0]->version);

#if 1 /* note, this isn't really necessary, so don't use for non-test applications */
	update_world();
	vrInputWaitForAllInputDevicesToBeOpen();
	vrFprintConfig(stdout, vrContext->config, brief);
	vrFprintInput(stdout, vrContext->input, verbose);
#endif

	vrDbgPrintf("VALTEST: looping\n");
	/* TODO: this really should use a queued event, because we often miss the key event */
	while(!vrGet2switchValue(0)) {
		update_world();
	}

	vrPrintf(BOLD_TEXT "VALTEST: I guess we're all done now.\n" NORM_TEXT);
	exit_application();

	exit(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application()
{
	vrDbgPrintfN(ALWAYS_DBGLVL, "VALTEST: quiting.\n");
	vrExit();
}

#ifdef PERSONAL_SIMULATOR
/********************************************************************/
void my_sim(vrRenderInfo *renderinfo, int mask)
{
#if 0
	vrPrintf("Yo, doing my own simulator!\n");
#endif
	vrRenderDefaultSimulator(renderinfo, mask);
}
#endif


/***********************/
/* update_world(): ... */
/***********************/
void update_world()
{
static	vrTime		lastTime = -1;
	vrTime		deltaTime;
	float		joy_x, joy_y;
	vrVector	vec;
	int		input;

	if (lastTime < 0)
		lastTime = vrCurrentSimTime();

	/* do some traveling. */
	deltaTime = vrCurrentSimTime() - lastTime;
	lastTime += deltaTime;

	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	if (fabs(joy_x) > 0.125) {
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, deltaTime * joy_x * -25.0);
	}

	if (fabs(joy_y) > 0.125) {
		vrVectorGetRWFrom6sensorDir(&vec, 1, VRVECTOR_FORE);
		vrVectorScale(&vec, deltaTime * joy_y * 2.5);
		vrUserTravelTranslateAd(VR_ALLUSERS, vec.v);
	}

	if (vrGet2switchValue(1)) {
		vrUserTravelReset(VR_ALLUSERS);
	}

	if (vrGet2switchValue(2)) {
		vrUserTravelTranslate3d(VR_ALLUSERS, 0.1, 0.0, 0.0);
	}
	if (vrGet2switchValue(3)) {
		vrUserTravelTranslate3d(VR_ALLUSERS, -0.1, 0.0, 0.0);
	}

	/* set DebugLevel to 200 to print this info to screen every frame */
	if (vrDbgDo(RARE_DBGLVL)) {
		vrMsgPrintf("VALTEST: controller --");

		vrMsgPrintf(" buttons:");
		for (input = 0; input < vrContext->input->num_2ways; input++)
			vrMsgPrintf(" %d", vrContext->input->switch2[input]->value);

		vrMsgPrintf(" valuators:");
		for (input = 0; input < vrContext->input->num_valuators; input++)
			vrMsgPrintf(" %6.3f", vrContext->input->valuator[input]->value);

		vrMsgPrintf("\n");
	}

	/* allow other processes to get some work done */
	vrSleep(10000);
}

