/* ======================================================================
 *
 *  CCCCC          vr_shmem.c
 * CC   CC         Author(s): Ed Peters, Bill Sherman, John Stone, Sukru Tikves
 * CC              Created: June 4, 1998
 * CC   CC         Last Modified: June 4, 2003
 *  CCCCC
 *
 * Code file for FreeVR shared memory.  Defines routines to initialize
 * and allocated from shared memory.  The pointers to the shared memory
 * arena are stored in static (file-scope) variables declared in this
 * file.  The file also contains definitions of the locking routines, as
 * well as the "private" locking structure.
 *
 * The code for initializing IRIX shared memory was mostly lifted from
 * Vijay Jaswal's cxMem.{h, C}, and Bill Sherman's shared memory code.
 *
 * Copyright 2003, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

TODO:
	- do more error checking and reporting using the errno facility.

	- probably should unify this with the regular sharedmemory
		routines, but it uses the shmctl/shmget/shmat/shmdt functions
		rather than the mmap/etc ones.  Getting input data out of shared
		memory requires some of these functions, however.

*************************************************************************/


/****************************************************************/
/* Some options to how the shared memory code shoud be compiled */
/*   (mostly options for use in debugging the library.)         */
#undef	VRDBXPAUSE_ON_ERROR	/* define to use debugger to analyze why no memory */
#undef	VRTRACKMEM		/* define this to keep track of memory allocations */
#undef	VRSHMEMINITDBG		/* define this to print debuging statements during initialization (ie. before Context is set) */
#define	VRTRACKLOCKS		/* define this to keep track of locking operations  -- DO NOT USE w/ vr_debug.c:PRINT_LOCK */
#undef	VRSHMEM_USESTRUCT	/* define this to keep all the shared mem info in a struct -- NYI */

#define VRSHMEM_ARENANAME "/tmp/freevr.arena"

#define	VRTRACE_SHMEM	/* define this to enable tracing within vrShmem...() calls */
#undef	VRTRACE_LOCK	/* define this to enable tracing within vrLock...() calls */
#define	VRTRACE_BARRIER	/* define this to enable tracing within vrBarrier...() calls */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#if defined(__sgi)
#  include <malloc.h> /* needed for amalloc/afree etc */
#endif


		/*********************************************/
		/*** Shared Memory structure and functions ***/
		/*********************************************/

#ifdef VRTRACE_SHMEM
#  define vrTraceOpt		vrTrace
#else
#  define vrTraceOpt(a,b)	;
#endif

#ifdef SHM_PF_ARENA /* use Performer's memory arena { */
#  include <Performer/pf.h>		/* needed for pfGetSharedArena() */
#  define PERFORMER_2_3_BUG  0		/* version 2.3 of Linux-Performer doesn't have pfGetSharedArena() */
#  if PERFORMER_2_3_BUG == 1
#    define amalloc		pfMalloc
#    define afree(ptr, ar)	pfFree((ptr))
#  endif
#else
#  define PERFORMER_2_3_BUG  0		/* When not using Performer, then we're not using pfGetSharedArena() */
#endif /* } */

#include "vr_debug.h"
#if 0 /* TODO: I'm not sure why Sukru included these */
#ifdef printf
#undef printf
#endif
#endif

#if defined(SHM_SYSVIPC) || defined(SEM_SYSVIPC)
#  include <sys/ipc.h>
#  include <sys/shm.h>
#  include <sys/sem.h>
#endif

#if defined(SEM_IRIX)
#  include <ulocks.h>
#endif

#if defined(SEM_POSIX)
#  include <semaphore.h>
#endif

#if defined(SEM_WIN32)
#  include <windows.h>
#endif

#if defined(SEM_TCP)
#  include "vr_sem_tcp.h"
#endif

#if defined(SHM_DUMMY)
   /* Stuart's simple wrappers to implement a dumb, but */
   /*   quick-to-write shared memory allocation scheme. */
   /*   (He wrote it in about 15 minutes.)              */

#  include <sys/mman.h>
#  ifndef SHM_PF_ARENA
     typedef void	usptr_t;
     typedef void	usema_t;
#    include "linux_dummy.c"
#  endif /* !SHM_PF_ARENA */
#  define MEM_SHARED 0x0001
#endif

#include "vr_shmem.h"
#include "vr_math.h"
#include "vr_context.h"

/***************************************************************/
/* File-scope variables for keeping track of the shared memory */
/* arena and whatnot.  Most of them are voodoo black-magic.    */
#ifdef VRTRACKMEM
#  define TRACK_SPACER	"                                                       "
#  ifndef VRSHMEM_USESTRUCT
   static	long		total_bytes_this_proc = 0;
   static	long		*total_bytes_overall = NULL;
   static	long		*total_freed_overall = NULL;
#  endif
#endif
#ifndef VRSHMEM_USESTRUCT
  static	long		total_bytes_too_many = 0;
  static	long		*arena_size = NULL;
#endif

#ifndef VRSHMEM_USESTRUCT
#  ifndef SHM_PF_ARENA
#    if defined(SHM_SVR4MMAP)
       static	long		shmem_fd = -1;
#    endif

     static	void		*shmem_addr = NULL;
     static	usptr_t		*usarena = NULL;
#    endif /* !SHM_PF_ARENA */
   static void *arena = NULL;

#  if defined(SHM_SYSVIPC)
     static	int		shmid = 0;
     static	int		shmkey = 0;
     static	struct shmid_ds	shminfo;
#  endif
#endif /* !VRSHMEM_USESTRUCT */

#ifdef VRSHMEM_USESTRUCT
  static	vrShmemInfo	shmem_info;
#endif


/*****************************************************************/
/*
 * this is a lot of black-magic calls to set up a shared memory
 * arena which will be shareable only by processes forked from
 * this one
 */
/* returns the size of the arena, which is especially useful if this */
/*   isn't the first call the vrShmemInit(), and thus the requested  */
/*   size won't match the actual size.                               */
long vrShmemInit(size_t bytes)
{
#ifndef SHM_PF_ARENA
	int	request_size;		/* used to adjust the bytes specified to be on a page boundary */
#endif

#ifdef VRSHMEMINITDBG
	vrTraceOpt("vrShmemInit", "beginning");
#endif

	if (arena) {
		vrErrPrintf("vrShmemInit(): " RED_TEXT "Warning: shared memory already initialized, additional attempt ignored\n" NORM_TEXT);
		return *arena_size;
	}

	/******************************************************/
	/* print some warnings on systems with known problems */
#if defined(SEM_POSIX) && !defined(_POSIX_THREAD_PROCESS_SHARED)
	vrMsgPrintf("FreeVR: " RED_TEXT 
			"WARNING, POSIX threads are not shared across processes on this system.\n"
		"        Features such as locking operations will not work.  Try compiling with\n"
		"        a different choice as the method of semaphore implementation.  See the\n"
		"        library development guide and the Make-arch file for details.\n" NORM_TEXT);
	vrSleep(2000000);
#endif
#if defined(SEM_SYSVIPC)
	vrMsgPrintf("FreeVR: " RED_TEXT 
			"WARNING, barriers based on SYS-V IPC semaphores are not fully functional.\n" NORM_TEXT);
	vrSleep(2000000);
#endif
	/* set the initial barrier values */

        /*********************************************************/
        /* create the address space for amalloc to allocate from */

	arena = NULL;

/***********************/
#ifdef SHM_PF_ARENA /* use Performer's memory arena { */
	/* NOTE: If a non-default arena size is desired, vrShmemInit() must be */
	/*   called before pfInit() -- otherwise the call to pfInitArenas()    */
	/*   will already have been made.                                      */
	pfSharedArenaSize(bytes);
	pfInitArenas();

	arena = pfGetSharedArena();

	/**************************************************************/
	/* The first thing in the arena will be the size of the arena */
	arena_size = (long *)pfMalloc(sizeof(long), arena);
	*arena_size = pfGetSharedArenaSize();

	vrNormPrintf("FreeVR: Just allocated an arena (0x%p) from Performer of size %ld\n", arena, *arena_size);

	if (arena == NULL) {
		vrNormPrintf(RED_TEXT "ERROR: if using Performer 2.3 on Linux, try recompiling library\n"
			"  with target \"linux-pf2.3\"\n" NORM_TEXT);
		exit(-1);
	}

#else /* } SHM_PF_ARENA vs. Use FreeVR Shared Memory Arena { */

        /* round request_size up to next full page of memory, required */
        /* by most mmap() shared memory implementations.               */
	request_size = ((bytes+getpagesize()-1) & ~(getpagesize()-1));

/***********************/
#if defined(SHM_SVR4MMAP)
        /* SVR4 style shared memory via mmap() of /dev/zero */
        shmem_fd = open("/dev/zero", O_RDWR );
        if (shmem_fd < 0) {
            vrErr("Couldn't open shmem file.");
            return 0;
        }
#  ifdef VRSHMEMINITDBG
        vrTraceOpt("vrShmemInit", "before mmap");
#  endif
#  if defined(__sgi)
        /* SGI's mmap lets us grow the mapped region on the fly if we like */
        shmem_addr = mmap(0, request_size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_AUTOGROW | MAP_AUTORESRV, shmem_fd, 0);
#  else
        /* non-SGI mmap doesn't */
        shmem_addr = mmap(0, request_size, PROT_READ | PROT_WRITE, MAP_SHARED, shmem_fd, 0);
#  endif

/***********************/
#elif defined(SHM_BSDANONMMAP)
        /* BSD4.4 style shared memory via mmap() and MAP_ANON */
        vrTraceOpt("vrShmemInit", "before mmap");
#  if defined(__hpux)
#     define MAP_ANON MAP_ANONYMOUS
#  endif
        shmem_addr = mmap(0, request_size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANON, -1, 0);

/***********************/
#elif defined(SHM_SYSVIPC)
        /* System V style shared memory via shmget() */
#  ifdef VRSHMEMINITDBG
        vrTraceOpt("vrShmemInit", "before shmget");
#  endif

	/* setup the shared memory, using the pid as the key */
        shmkey = getpid(); /* TODO: maybe this should be configurable */
        shmid = shmget(shmkey, request_size, IPC_CREAT | SHM_R |SHM_W);
	if (shmid < 0) {
		switch (errno) {
		case ENOENT:
			/* NOTE: actually this shouldn't happen since we're creating the arena */
			vrErrPrintf("vrShmemInit(): " RED_TEXT "ERROR: no such shared memory key\n" NORM_TEXT);
			break;
		case EINVAL:
			/* outside the bounds of possible shared memory. */
			vrErrPrintf("vrShmemInit(): " RED_TEXT "ERROR: outside memory bounds\n" NORM_TEXT);
			break;
		case ENOMEM:
			/* not enough shared memory available */
			vrErrPrintf("vrShmemInit(): " RED_TEXT "ERROR: not enough memory available\n" NORM_TEXT);
			break;
		case ENOSPC:
			/* this only occurs when creating new keys -- not */
			/*   enough system resources for another key.     */
			vrErrPrintf("vrShmemInit(): " RED_TEXT "ERROR: no more shared memory segments allowed\n" NORM_TEXT);
			break;
		case EEXIST:
			/* a shared memory segment with that key already exists
	*/
			vrErrPrintf("vrShmemInit(): " RED_TEXT "ERROR: key %x already exists\n" NORM_TEXT, shmkey);
			break;

		}

		return 0;
	}

	/* set the permissions for the arean */
        shmctl(shmid, IPC_STAT, &shminfo);
        shminfo.shm_perm.mode = 0666;
        shmctl(shmid, IPC_SET, &shminfo);

	/* now get the shared memory address */
        shmem_addr = shmat(shmid, NULL, SHM_RND | SHM_R | SHM_W);
        if (shmem_addr == NULL || shmem_addr == (void *)-1) {
		shmctl(shmid, IPC_RMID, &shminfo);
        }
#endif

        if (shmem_addr == NULL || shmem_addr == (void *)-1) {
                vrErr("Couldn't acquire shared memory from system.");
                return 0;
        }


	/*************************************/
	/*** Now initialize the semaphores ***/
	/*************************************/

/***********************/
#if defined(SEM_IRIX) || (defined(SHM_SVR4MMAP) && defined(__sgi)) /* { */
        /* This code is only functional on IRIX, emulated everywhere else */
        /* So this initialization sequence doesn't make sense elsewhere   */
	/*****************************************/
	/* create an arena to be used by acreate */
#  ifdef VRSHMEMINITDBG
	vrTraceOpt("vrShmemInit", "before usinit");
#  endif
	usarena = usinit(VRSHMEM_ARENANAME);
	if (!usarena) {
		vrErr("Couldn't init arena.");
		return 0;
	}
/***********************/
#else /* } { */
        usarena = NULL; /* unused, but passed into emulation code */
#endif /* } */

	/**********************************************************/
	/* Remove the arena file, so only process descending from */
	/* this one can connect to our shared memory arena.       */
#ifdef VRSHMEMINITDBG
	vrTraceOpt("vrShmemInit", "before arena file unlink");
#endif
	unlink(VRSHMEM_ARENANAME);

	/************************************/
	/* now do the actual arena creation */
#ifdef VRSHMEMINITDBG
	vrTraceOpt("vrShmemInit", "before acreate");
#endif
	arena = acreate(shmem_addr, bytes, MEM_SHARED, usarena, NULL);
	if (!arena) {
		vrErr("Couldn't create arena.");
		return 0;
	}

	/**************************************************************/
	/* The first thing in the arena will be the size of the arena */
	arena_size = (long *)amalloc(sizeof(long), arena);
	*arena_size = bytes;
#endif /* } !SHM_PF_ARENA */ /* TODO: should this come after the VRTRACKMEM stuff? */

#ifdef VRTRACKMEM
	if (arena || PERFORMER_2_3_BUG) {
		total_bytes_overall = (long *)amalloc(sizeof(long), arena);
		total_freed_overall = (long *)amalloc(sizeof(long), arena);
	} else	vrErr("No shmem arena to allocate from.");

	if (total_bytes_overall == NULL) {
		total_bytes_too_many += sizeof(long);
		vrErrPrintf("vrShmemInit(%d): "
			RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
			bytes, total_bytes_too_many);
#  ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#  endif
	}

	*total_bytes_overall = 3 * sizeof(long);	/* 3 longs allocated so far */
	*total_freed_overall = 0;
	total_bytes_this_proc = *total_bytes_overall;
	vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
#endif  /* VRTRACKMEM */
#ifdef VRSHMEMINITDBG
	vrTraceOpt("vrShmemInit", "done");
#endif

	return *arena_size;
}


/*****************************************************************/
/*
 * Detach shared memory resources, delete any allocated shared memory
 * keys, file handles, etc so that we don't unintentionally consume
 * finite system resources (such as keys, files, etc that last beyond
 * process exit) with each run.
 */
void vrShmemExit()
{
#if defined(SHM_SVR4MMAP) || defined(SHM_BSDANONMMAP)
	int	size = *arena_size;
#endif

#ifndef SHM_PF_ARENA /* if using Performer's memory system, let it clean itself up { */
	adelete(arena);
	usdetach(usarena);
#  if defined(SHM_SVR4MMAP)
	munmap(shmem_addr, size);		/* detach shared memory          */
	close(shmem_fd);			/* close memory mapped /dev/zero */
#  elif defined(SHM_BSDANONMMAP)
	munmap(shmem_addr, size);		/* detach shared memory          */
#  elif defined(SHM_SYSVIPC)
        shmdt(shmem_addr);			/* detach shared memory          */
        shmctl(shmid, IPC_RMID, &shminfo);	/* delete shared memory key      */

	shmid = 0;
	shmkey = 0;
#    if 0 /* this isn't a pointer, so setting to NULL isn't correct */
	shminfo = NULL;
#    endif
#  endif
	shmem_addr = NULL;
	usarena = NULL;
#endif /* } !SHM_PF_ARENA */
	arena = NULL;		/* after this FreeVR will no longer be able to use shared memory */
}


/*****************************************************************/
void *vrShmemArena()
{
	return arena;
}


/*****************************************************************/
long vrShmemUsage()
{
#ifdef VRTRACKMEM
	return *total_bytes_overall;
#else
	return -1L;
#endif
}


/*****************************************************************/
long vrShmemFreed()
{
#ifdef VRTRACKMEM
	return *total_freed_overall;
#else
	return -1L;
#endif
}


/*****************************************************************/
/*
 * Once the arena has been initialized, these are the simplest
 * functions in the world (just wrappers around amalloc and afree).
 */
void *vrShmemAlloc(size_t bytes)
{
	void	*mem = NULL;

	if (bytes == 0) {
		vrDbg("Requested allocation of 0 bytes.");
		return mem;
	}

	if (arena || PERFORMER_2_3_BUG)
		mem = amalloc(bytes, arena);
	else	vrErr("No shmem arena to allocate from.");

	if (mem == NULL) {
		total_bytes_too_many += bytes;
		vrErrPrintf("vrShmemAlloc(%d): "
			RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
			bytes, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
	}

#ifdef VRTRACKMEM
	if (mem) {
		*total_bytes_overall += amallocblksize(mem, arena);
		total_bytes_this_proc += amallocblksize(mem, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif

	return mem;
}


/*****************************************************************/
void *vrShmemAlloc0(size_t bytes)
{
	void	*mem = NULL;

	if (bytes == 0) {
		vrDbg("Requested allocation of 0 bytes.");
		return mem;
	}

	if (arena || PERFORMER_2_3_BUG) {
		mem = amalloc(bytes, arena);
	} else	vrErr("No shmem arena to allocate from.");

	if (mem == NULL) {
		total_bytes_too_many += bytes;
		vrErrPrintf("vrShmemAlloc0(%d): "
			RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
			bytes, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
	} else {
		memset(mem, 0, bytes);
	}

#ifdef VRTRACKMEM
	if (mem) {
		*total_bytes_overall += amallocblksize(mem, arena);
		total_bytes_this_proc += amallocblksize(mem, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif

	return mem;
}


/*****************************************************************/
void *vrShmemRealloc(void *mem, size_t bytes)
{
#ifdef VRTRACKMEM
	long old_size = (mem ? amallocblksize(mem, arena) : 0);
#endif

	if (arena || PERFORMER_2_3_BUG)
		mem = arealloc(mem, bytes, arena);
	else	vrErr("No shmem arena to reallocate from.");

	if (mem == NULL) {
		total_bytes_too_many += bytes;
		vrErrPrintf("vrShmemRealloc(%d): "
			RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
			bytes, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
	}

#ifdef VRTRACKMEM
	if (mem) {
		*total_bytes_overall += (amallocblksize(mem, arena) - old_size);
		total_bytes_this_proc += (amallocblksize(mem, arena) - old_size);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif

	return mem;
}


/*****************************************************************/
void vrShmemFree(void *mem)
{
#ifdef VRTRACKMEM
	if (mem) {
		*total_bytes_overall -= amallocblksize(mem, arena);
		*total_freed_overall += amallocblksize(mem, arena);
		total_bytes_this_proc -= amallocblksize(mem, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far -- freed %d\r", *total_bytes_overall, *total_freed_overall);
	}
#endif
	if ((arena || PERFORMER_2_3_BUG) && mem)
		afree(mem, arena);
}



/*****************************************************************/
/* Some functions for creating memory for specific types of things */


/*****************************************************************/
char *vrShmemStrDup(char *s)
{
	char	*buf = NULL;

	if (arena || PERFORMER_2_3_BUG) {
		buf = (char *)amalloc(strlen(s) + 1, arena);
		if (buf == NULL) {
			total_bytes_too_many += strlen(s) + 1;
			vrErrPrintf("vrShmemStrDup(%s): "
				RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
				s, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
			return(buf);
		}

		strcpy(buf, s);
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (buf) {
		*total_bytes_overall += amallocblksize(buf, arena);
		total_bytes_this_proc += amallocblksize(buf, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return buf;
}


/*****************************************************************/
char *vrShmemStrCat(char *s1, char *s2)
{
	char	*buf = NULL;

	if (arena || PERFORMER_2_3_BUG) {
		buf = (char *)amalloc(strlen(s1) + strlen(s2) + 1, arena);
		if (buf == NULL) {
			total_bytes_too_many += strlen(s1) + strlen(s2) + 1;
			vrErrPrintf("vrShmemStrCat(%s, %s): "
				RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
				s1, s2, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
			return(buf);
		}

		strcpy(buf, s1);
		strcpy(buf+strlen(s1), s2);
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (buf) {
		*total_bytes_overall += amallocblksize(buf, arena);
		total_bytes_this_proc += amallocblksize(buf, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return buf;
}


/*****************************************************************/
void *vrShmemMemDup(void *mem, int size)
{
	void	*buf = NULL;

	if (mem == NULL)
		return NULL;

	if (arena || PERFORMER_2_3_BUG) {
		buf = (void *)amalloc(size, arena);
		if (buf == NULL) {
			total_bytes_too_many += size;
			vrErrPrintf("vrShmemMemDup(0x%p, %d): "
				RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
				mem, size, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
			return(buf);
		}

		memcpy(buf, mem, size);
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (buf) {
		*total_bytes_overall += amallocblksize(buf, arena);
		total_bytes_this_proc += amallocblksize(buf, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return buf;
}


/*****************************************************************/
/* 6/12/01: this function currently not used -- was used in old config method */
int *vrShmemIntArray(int count,...)
{
	int	*ary = NULL;

	if (arena || PERFORMER_2_3_BUG) {
		if (count) {
			int i = 0;
			va_list ap;
			ary = (int *)amalloc(count * sizeof(int), arena);
			if (ary == NULL) {
				total_bytes_too_many += count * sizeof(int);
				vrErrPrintf("vrShmemIntArray(%d, ...): "
					RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
					count, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
				return(ary);
			}

			va_start(ap, count);
			for (i = 0; i < count; i++)
				ary[i] = va_arg(ap, int);

			va_end(ap);
		}
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (ary) {
		*total_bytes_overall += amallocblksize(ary, arena);
		total_bytes_this_proc += amallocblksize(ary, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return ary;
}


/*****************************************************************/
/* 6/12/01: this function currently not used -- was used in old config method */
float *vrShmemFloatArray(int count,...)
{
	float	*ary = NULL;

	if (arena || PERFORMER_2_3_BUG) {
		if (count) {
			int i = 0;
			va_list ap;
			ary = (float *)vrShmemAlloc(count * sizeof(float));
			if (ary == NULL) {
				total_bytes_too_many += count * sizeof(float);
				vrErrPrintf("vrShmemFloatArray(%d, ...): "
					RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
					count, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
				return(ary);
			}

			va_start(ap, count);
			for (i = 0; i < count; i++)
                                ary[i] = va_arg(ap, double); /* the va_arg _must_ be double, even though */
                                                             /* we're working with floats.               */

			va_end(ap);
		}
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (ary) {
		*total_bytes_overall += amallocblksize(ary, arena);
		total_bytes_this_proc += amallocblksize(ary, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return ary;
}


/*****************************************************************/
/* 6/12/01: this function currently not used -- was used in old config method */
void **vrShmemPtrArray(int count,...)
{
	void	**ary = NULL;

	if (arena || PERFORMER_2_3_BUG) {
		if (count) {
			int i = 0;
			va_list ap;
			ary = (void **)vrShmemAlloc(count * sizeof(void *));
			if (ary == NULL) {
				total_bytes_too_many += count * sizeof(void *);
				vrErrPrintf("vrShmemPtrArray(%d, ...): "
					RED_TEXT "No shmem allocated!  IE. out of shared memory. (%ld)\n" NORM_TEXT,
					count, total_bytes_too_many);
#ifdef VRDBXPAUSE_ON_ERROR
		vrErrPrintf(BOLD_TEXT "pausing, use 'dbx -p %d' to debug\n" NORM_TEXT, getpid());
		pause();
#endif
				return(ary);
			}

			va_start(ap, count);
			for (i = 0; i < count; i++)
				ary[i] = va_arg(ap, void *);

			va_end(ap);
		}
	} else {
		vrErr("No shmem arena to allocate from.");
	}

#ifdef VRTRACKMEM
	if (ary) {
		*total_bytes_overall += amallocblksize(ary, arena);
		total_bytes_this_proc += amallocblksize(ary, arena);
		vrDbgPrintfN(TRACKMEM_DBGLVL, TRACK_SPACER "(shmem) allocated %d bytes so far\r", *total_bytes_overall);
	}
#endif
	return ary;
}



		/********************************************/
		/*** Data Locking structure and functions ***/
		/********************************************/

/*
 * locking primitives --
 * for the most part, this code just apes stuff I (elp) found in the
 * dinosaur book (Silberschatz and Galvin's "Operating System
 * Concepts", 4th edition, pp 183-4)
 */

#undef vrTraceOpt
#ifdef VRTRACE_LOCK
#  define vrTraceOpt		vrTrace
#else
#  define vrTraceOpt(a,b)	;
#endif

/*************************/
#if defined(SEM_SYSVIPC)
	/* may have to be defined manually on some systems */

   /***************/
#  if defined(__sgi) || defined(__OpenBSD__) || defined(__NetBSD__)
union semun semarg;
#  elif defined(__sun) || defined(_AIX)
union semun {
		int             val;
		struct semid_ds *buf;
		ushort_t        *array;
	} semarg;

   /***************/
#  elif defined(__linux)
union semun {
		int		val;		/* value for SETVAL */
		struct semid_ds	*buf;		/* buffer for IPC_STAT, IPC_SET */
	unsigned short int	*array;		/* array for GETALL, SETALL */
		struct seminfo	*__buf;		/* buffer for IPC_INFO */
	} semarg;

   /***************/
#  elif defined(__hpux)
union semun {
		int             val;
		struct semid_ds *buf;
	} semarg;
#  endif

#  if !defined(SEM_R)
#     define SEM_R 0400
#  endif

#  if !defined(SEM_A)
#     define SEM_A 0200
#  endif
#endif


/*************************************************/
/* NOTE: the public vrLock datatype is just a (void *) type, */
/*   which can then be passed to the routines here, which    */
/*   treat the type as this fullly-blown defined structure.  */
typedef struct vrLock_st {
		/***************************/
		/* Generic Object field(s) */
		vrObject	object_type;	/* The type of this object (ie VROBJECT_LOCK) */
		/* NOTE: the rest of the generic object fields not included with this struct  */
		/* TODO: consider adding the other object fields to this structure.           */

		/***************************/
		/* specific lock fields */
		char		*name;		/* name to use with debugging output */
	struct	vrLock_st	*next;		/* pointer to the next lock in the list */
		int		trace;		/* flag to do tracing debug output */
		int		readcount;	/* counter of number of readers locking this lock */
		int		total_readsets;	/* counter of total number read-sets on this lock */
		int		total_readrels;	/* counter of total number read-releases on this lock */
		int		total_writesets;/* counter of total number write-sets on this lock */
		int		total_writerels;/* counter of total number write-releases on this lock */

/***********************/
#if defined(SEM_IRIX)
		usema_t		*wmutex;	/* semaphore ID of write semaphore */
		usema_t		*rmutex;	/* semaphore ID of readcount access semaphore */

/***********************/
#elif defined(SEM_POSIX)
		sem_t		*wmutex;	/* semaphore ID of write semaphore */
		sem_t		*rmutex;	/* semaphore ID of readcount access semaphore */

/***********************/
#elif defined(SEM_SYSVIPC)
		struct sembuf	waitop;		/* wait operation */
		struct sembuf	postop;		/* post operation */
                int		wmutex;		/* semaphore ID of write semaphore */
                int		rmutex;		/* semaphore ID of readcount access semaphore */

/***********************/
#elif defined(SEM_WIN32)
		HANDLE		wmutex;		/* semaphore ID of write semaphore */
		HANDLE		rmutex;		/* semaphore ID of readcount access semaphore */

/***********************/
#elif defined(SEM_TCP)
		MUTEX_HANDLE	wmutex;		/* semaphore ID of write semaphore */
		MUTEX_HANDLE	rmutex;		/* semaphore ID of readcount access semaphore */

/***********************/
#else
		/* TODO: is this good enough, or is there more? */
		void		*wmutex;	/* semaphore ID of write semaphore */
		void		*rmutex;	/* semaphore ID of readcount access semaphore */
#endif
	} vrPrivateLock;


/*****************************************************************/
/* This is just for version 0.4f, since I've changed the arguments for  */
/*   for vrCreateLock(), and will be making more changes in version 0.5 */
vrLock vrCreateLockOld()
{
	return vrCreateLock(vrContext);
}


/*****************************************************************/
/*
 * the private lock structure will have two semaphores -- one
 * to control access to a count of readers and one to control
 * write access (only one writer allowed at a time)
 */
/* TODO: I don't know why we're forced to use "void *" here, when it works fine for vrBarrierCreate() -- figure this out! */
vrLock vrCreateLock(void *void_context)
{
	vrContextInfo		*context = (vrContextInfo *)void_context;
	vrPrivateLock		*lock = (vrPrivateLock *)vrShmemAlloc(sizeof(vrPrivateLock));
#if (defined(SEM_POSIX) && defined(__linux)) || defined(SEM_SYSVIPC)
	int			werrno = 0;	/* errno from an operation on the write semaphore */
	int			rerrno = 0;	/* errno from an operation on the read semaphore */
#endif
#if defined(SEM_WIN32)
	SECURITY_ATTRIBUTES	sec;
#endif

	/**************************************/
	/* add this lock to the list of locks */
	if (context != NULL) {
		vrSetWriteLock(context->head_lock);	/* the head_lock is the lock for the lock list */
		((vrPrivateLock *)(context->tail_lock))->next = lock;
		context->tail_lock = lock;
		vrReleaseWriteLock(context->head_lock);
	}

	/******************************/
	/* assign some initial values */
	lock->object_type = VROBJECT_LOCK;
	lock->trace = 0;
	lock->readcount = 0;
	lock->total_readsets = 0;
	lock->total_readrels = 0;
	lock->total_writesets = 0;
	lock->total_writerels = 0;

	/***********************************************************/
	/* setup the mutual-exclusion semaphores (system dependent) */

/***********************/
#if defined(SEM_IRIX)
	lock->wmutex = usnewsema(usarena, 1);
	lock->rmutex = usnewsema(usarena, 1);

/***********************/
#elif defined(SEM_POSIX)
        lock->wmutex = vrShmemAlloc(sizeof(sem_t));
        lock->rmutex = vrShmemAlloc(sizeof(sem_t));
#  if !defined(_POSIX_THREAD_PROCESS_SHARED)
        if (sem_init(lock->wmutex, 0, 1) < 0)
		werrno = errno;
        if (sem_init(lock->rmutex, 0, 1) < 0)
		rerrno = errno;

	vrDbgPrintfN(AALWAYS_DBGLVL, "SEM_POSIX sem_init errnos: %d %d\n", werrno, rerrno);
#  else
        sem_init(lock->wmutex, 1, 1);
        sem_init(lock->rmutex, 1, 1);
#  endif

/***********************/
#elif defined(SEM_SYSVIPC)
        lock->waitop.sem_num = 0;
        lock->waitop.sem_op  = -1;
        lock->waitop.sem_flg = 0;
        lock->postop.sem_num = 0;
        lock->postop.sem_op  = 1;
        lock->postop.sem_flg = 0;
	errno = 0;
        lock->wmutex = semget(IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | SEM_A | SEM_R);
	werrno = errno;
        lock->rmutex = semget(IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | SEM_A | SEM_R);
	rerrno = errno;
        if ((lock->wmutex < 0) || (lock->rmutex < 0)) {
		/* failed to acquire semaphores */
		vrErrPrintf("vrCreateLock(): " RED_TEXT "Can't get the necessary semaphore (%d, %d).\n" NORM_TEXT, lock, werrno, rerrno);
		perror("vrCreateLock");
		context->startup_error |= VRSTARTUP_NOLOCKSEM;
        } else {
		/* initialize SYSV semaphores, this is no fun... */
		semarg.val = 0;
		semctl(lock->wmutex, 0, SETVAL, semarg);
		semctl(lock->rmutex, 0, SETVAL, semarg);
		semop(lock->wmutex, &lock->postop, 1);
		semop(lock->rmutex, &lock->postop, 1);
	}

/***********************/
#elif defined(SEM_WIN32)
	/* set the security attributes */
	memset(&sec, 0, sizeof(sec));
	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = 1;

	lock->wmutex = CreateSemaphore(&sec, 1, 1, NULL);
	lock->rmutex = CreateSemaphore(&sec, 1, 1, NULL);

/***********************/
#elif defined(SEM_TCP)
	lock->wmutex = vrTcpNewMutex();
	lock->rmutex = vrTcpNewMutex();

/***********************/
#else
	/* There is no default lock scheme, so choose one of the above, or make a new one. */
	/* Without more information, basically make a nop lock */
	lock->wmutex = NULL;
	lock->rmutex = NULL;
#endif

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) created lock %p\n", lock);
#endif

	return (vrLock)lock;
}


/*****************************************************************/
vrLock vrLockCreateName(void *context, char *name)
{
	vrLock		lock = vrCreateLock(context);
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (name && *name)
		plock->name = vrShmemStrDup(name);
	else	plock->name = "";

	return lock;
}


/*****************************************************************/
vrLock vrCreateTracedLock(void *context, char *name)
{
static	char		string[128];
	vrLock		lock = vrCreateLock(context);
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	plock->trace = 1;
	if (name && *name)
		plock->name = vrShmemStrDup(name);
	else	plock->name = "";

	vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Tracing new lock '%s' [%p], status %s\n",
			plock->name, plock, vrLockStatus(lock, string));
	vrFprintLock(stdout, plock, verbose);

	return lock;
}


/*****************************************************************/
/*
 * a freed lock will have both semaphores set to NULL;
 * it's never an error to release a lock
 */
void vrLockFree(vrLock lock)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (plock->trace)
		vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Deleting lock '%s' [%p], pre-status %s\n",
				plock->name, plock, vrLockStatus(lock, string));

	vrReleaseWriteLock(lock);
	vrReleaseReadLock(lock);
	vrSetWriteLock(lock);

/***********************/
#if defined(SEM_IRIX)
	usfreesema(plock->wmutex, usarena);
	usfreesema(plock->rmutex, usarena);
	plock->wmutex = plock->rmutex = NULL;

/***********************/
#elif defined(SEM_POSIX)
        sem_destroy(plock->wmutex);
        sem_destroy(plock->rmutex);
        vrShmemFree(plock->wmutex);
        vrShmemFree(plock->rmutex);
	plock->wmutex = plock->rmutex = NULL;

/***********************/
#elif defined(SEM_SYSVIPC)
        semctl(plock->wmutex, 0, IPC_RMID);
        semctl(plock->rmutex, 0, IPC_RMID);
	plock->wmutex = plock->rmutex = -1;

/***********************/
#elif defined(SEM_WIN32)
	CloseHandle(plock->wmutex);
	CloseHandle(plock->rmutex);
	plock->wmutex = plock->rmutex = NULL;

/***********************/
#elif defined(SEM_TCP)
	vrTcpDeleteMutex(plock->wmutex);
	vrTcpDeleteMutex(plock->rmutex);

/***********************/
#else
	/* There is no default lock scheme, so choose one of the above, or make a new one. */
#endif

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) freed lock %p\n", lock);
#endif
	return;
}


/*****************************************************************/
/* Loop over the linked list, calling vrLockFree() for each */
void vrLockFreeAll(vrLock lock)
{
	vrPrivateLock	*first = (vrPrivateLock *)lock;
	vrPrivateLock	*plock;
	vrPrivateLock	*next;

	for (plock = first; plock != NULL; plock = next) {
		next = plock->next;
		vrLockFree(plock);
	}
}


/********************************/
/* Verify that "lock" is a lock */
int vrLockCheck(vrLock lock)
{
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (plock == NULL)
		return 0;

	return (plock->object_type == VROBJECT_LOCK);
}


/*****************************************************************/
/*
 * a new reader will increment readcount
 * if readcount was previously 0, we'll want to get a write lock as well
 *
 * a finished reader will decrement the readcount
 * if readcount is decremented to 0, then release the writelock
 */
void vrSetReadLock(vrLock lock)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (plock->trace)
		vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Acquiring read lock '%s' [%p], pre-status %s\n",
				plock->name, plock, vrLockStatus(lock, string));

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) setting lock %p for reading\n", lock);
#endif

	if (!plock->rmutex) {
		vrErrPrintf("vrSetReadLock(): " RED_TEXT "Can't set read lock: bad lock [%p].\n" NORM_TEXT, lock);

#ifdef VRDBXPAUSE_ON_ERROR
		vrPrintf(BOLD_TEXT "use 'dbx -p %d' to debug\n", getpid());
		pause();
#endif
		return;
	}

	plock->total_readsets++;
#ifdef VRTRACE_LOCK
	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s '%s'\n", __FILE__, __LINE__, "vrSetReadLock", "about to lock", plock->name);
#endif

/***********************/
#if defined(SEM_IRIX)
	uspsema(plock->rmutex);					/* set lock on readcount data */
	plock->readcount++;
	if (plock->readcount == 1)				/* first reader sets exclusion lock */
		uspsema(plock->wmutex);
	usvsema(plock->rmutex);					/* release lock on readcount data */

/***********************/
#elif defined(SEM_POSIX)
        sem_wait(plock->rmutex);				/* set lock on readcount data */
	plock->readcount++;
	if (plock->readcount == 1)				/* first reader sets exclusion lock */
		sem_wait(plock->wmutex);
        sem_post(plock->rmutex);				/* release lock on readcount data */

/***********************/
#elif defined(SEM_WIN32)
	_WaitForSemaphore(plock->rmutex);			/* set lock on readcount data */
	plock->readcount++;
	if (plock->readcount == 1)				/* first reader sets exclusion lock */
		_WaitForSemaphore(plock->wmutex);
	ReleaseSemaphore(plock->rmutex, 1, NULL);		/* release lock on readcount data */

/***********************/
#elif defined(SEM_SYSVIPC)
        semop(plock->rmutex, &plock->waitop, 1);		/* set lock on readcount data */
	plock->readcount++;
	if (plock->readcount == 1)				/* first reader sets exclusion lock */
		semop(plock->wmutex, &plock->waitop, 1);
        semop(plock->rmutex, &plock->postop, 1);		/* release lock on readcount data */

/***********************/
#elif defined(SEM_TCP)
	vrTcpAcquireMutex(plock->rmutex);			/* set lock on readcount data */
	plock->readcount++;
	if (plock->readcount == 1)				/* first reader sets exclusion lock */
		vrTcpAcquireMutex(plock->wmutex);
	vrTcpReleaseMutex(plock->rmutex);			/* release lock on readcount data */

/***********************/
#else
	/* TODO: No default lock scheme -- should we print a warning? */
#endif
	vrTraceOpt("vrSetReadLock", "locked");
}


/*****************************************************************/
void vrReleaseReadLock(vrLock lock)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (plock->trace)
		vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Releasing read lock '%s' [%p], pre-status %s\n",
				plock->name, plock, vrLockStatus(lock, string));

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) freeing lock %p for reading\n", lock);
#endif

	if (!plock->rmutex)
		return;

	plock->total_readrels++;
#ifdef VRTRACE_LOCK
	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s '%s'\n", __FILE__, __LINE__, "vrReleaseReadLock", "about to release", plock->name);
#endif

/***********************/
#if defined(SEM_IRIX)
	uspsema(plock->rmutex);					/* set lock on readcount data */
	plock->readcount--;
	if (plock->readcount == 0)				/* last readlock releases exclusion lock */
		usvsema(plock->wmutex);
	usvsema(plock->rmutex);					/* release lock on readcount data */

/***********************/
#elif defined(SEM_POSIX)
	sem_wait(plock->rmutex);				/* set lock on readcount data */
	plock->readcount--;
	if (plock->readcount == 0)				/* last readlock releases exclusion lock */
		sem_post(plock->wmutex);
	sem_post(plock->rmutex);				/* release lock on readcount data */

/***********************/
#elif defined(SEM_WIN32)
	_WaitForSemaphore(plock->rmutex);			/* set lock on readcount data */
	plock->readcount--;
	if (plock->readcount == 0)				/* last readlock releases exclusion lock */
		ReleaseSemaphore(plock->wmutex, 1, NULL);
	ReleaseSemaphore(plock->rmutex, 1, NULL);		/* release lock on readcount data */

/***********************/
#elif defined(SEM_SYSVIPC)
	semop(plock->rmutex, &plock->waitop, 1);		/* set lock on readcount data */
	plock->readcount--;
	if (plock->readcount == 0)				/* last readlock releases exclusion lock */
		semop(plock->wmutex, &plock->postop, 1);
        semop(plock->rmutex, &plock->postop, 1);		/* release lock on readcount data */

/***********************/
#elif defined(SEM_TCP)
	vrTcpAcquireMutex(plock->rmutex);			/* set lock on readcount data */
	plock->readcount--;
	if (plock->readcount == 0)				/* last readlock releases exclusion lock */
		vrTcpReleaseMutex(plock->wmutex);
	vrTcpReleaseMutex(plock->rmutex);			/* release lock on readcount data */

/***********************/
#else
	/* TODO: No default lock scheme -- should we print a warning? */
#endif

	vrTraceOpt("vrReleaseReadLock", "released");
}


/*****************************************************************/
void vrReadLockToWriteLock(vrLock lock)
{
	vrReleaseReadLock(lock);
	vrSetWriteLock(lock);
}


/*****************************************************************/
/*
 * a new writer waits to acquire the write lock
 * and frees it when they're done
 */
void vrSetWriteLock(vrLock lock)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (plock->trace)
		vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Acquiring write lock '%s' [%p], pre-status %s\n",
				plock->name, plock, vrLockStatus(lock, string));

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) setting lock %p for writing\n", lock);
#endif

	if (!plock->wmutex) {
		vrErrPrintf("vrSetWriteLock(): " RED_TEXT "Can't set write lock: bad lock [%p].\n" NORM_TEXT, lock);
		return;
	}

	plock->total_writesets++;
#ifdef VRTRACE_LOCK
	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s '%s'\n", __FILE__, __LINE__, "vrSetWriteLock", "about to lock", plock->name);
#endif

/***********************/
#if defined(SEM_IRIX)
	uspsema(plock->wmutex);

/***********************/
#elif defined(SEM_POSIX)
	sem_wait(plock->wmutex);

/***********************/
#elif defined(SEM_WIN32)
	_WaitForSemaphore(plock->wmutex);

/***********************/
#elif defined(SEM_SYSVIPC)
	semop(plock->wmutex, &plock->waitop, 1);

/***********************/
#elif defined(SEM_TCP)
	vrTcpAcquireMutex(plock->wmutex);

/***********************/
#else
	/* TODO: No default lock scheme -- should we print a warning? */
#endif

	vrTraceOpt("vrSetWriteLock", "locked");
}


/*****************************************************************/
void vrReleaseWriteLock(vrLock lock)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if(plock->trace)
		vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Releasing write lock '%s' [%p], pre-status %s\n",
				plock->name, plock, vrLockStatus(lock, string));

#ifdef VRTRACKLOCKS
	vrDbgPrintfN(TRACKLOCKS_DBGLVL, "(lock) freeing lock %p for writing\n", lock);
#endif

	if (!plock->wmutex) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "vrReleaseWriteLock(): " RED_TEXT "Something is wrong with the lock's wmutex\n" NORM_TEXT);
		return;
	}

	plock->total_writerels++;
#ifdef VRTRACE_LOCK
	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s '%s'\n", __FILE__, __LINE__, "vrReleaseWriteLock", "about to release", plock->name);
#endif

/***********************/
#if defined(SEM_IRIX)
	usvsema(plock->wmutex);

/***********************/
#elif defined(SEM_POSIX)
	sem_post(plock->wmutex);

/***********************/
#elif defined(SEM_WIN32)
	ReleaseSemaphore(plock->wmutex, 1, NULL);

/***********************/
#elif defined(SEM_SYSVIPC)
	semop(plock->wmutex, &plock->postop, 1);

/***********************/
#elif defined(SEM_TCP)
	vrTcpReleaseMutex(plock->wmutex);

/***********************/
#else
	/* TODO: No default lock scheme -- should we print a warning? */
#endif

	vrTraceOpt("vrReleaseWriteLock", "released");
}


/*****************************************************************/
void vrTraceLock(vrLock lock, int trace)
{
static	char		string[128];
	vrPrivateLock	*plock = (vrPrivateLock *)lock;

	if (vrLockCheck(plock)) {
		plock->trace = trace;
	}

	vrDbgPrintfN(ALWAYS_DBGLVL, "LOCK: Tracing lock '%s' [%p] now " BOLD_TEXT "%s" NORM_TEXT ", pre-status %s\n",
		plock->name, plock, (trace ? "on" : "off"), vrLockStatus(lock, string));
}


/*****************************************************************/
/* vrLockStatus() returns a string with a summary of the lock's  */
/*   current status.  It is the same format at the one_line print*/
/*   style status given for a lock.  This function is intended   */
/*   for use by the lock-tracing routines.                       */
/* NOTE: string should have memory allocated, which is where the */
/*   information will be placed.                                 */
char *vrLockStatus(vrLock lock, char *string)
{
	vrPrivateLock	*plock = (vrPrivateLock *)lock;
	int		wvalue;
	int		rvalue;
#if defined(SEM_SYSVIPC)
	int		wncnt, rncnt;
#endif
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
	int		lock_held, read_held, write_held;
	int		read_procs_holding;
	int		procs_waiting;
	int		read_procs_waiting;
	int		write_procs_waiting;
#endif

	/* if null pointer given, print an empty shell and return */
	if (plock == NULL) {
		sprintf(string, "<nil> lock");
		return string;
	}

	/****************************************************************/
	/* Get the values of the semaphores based on the defined system */
/***********************/
#if defined(SEM_IRIX)
		wvalue = ustestsema(plock->wmutex);
		rvalue = ustestsema(plock->rmutex);

/***********************/
#elif defined(SEM_POSIX)
		sem_getvalue(plock->wmutex, &wvalue);
		sem_getvalue(plock->rmutex, &rvalue);

/***********************/
#elif defined(SEM_SYSVIPC)
		wvalue = semctl(plock->wmutex, 0, GETVAL);
		wncnt  = semctl(plock->wmutex, 0, GETNCNT);
		rvalue = semctl(plock->rmutex, 0, GETVAL);
		rncnt  = semctl(plock->rmutex, 0, GETNCNT);
/***********************/
#elif defined(SEM_WIN32) && 0 /* disabled until the correct code is written */

/***********************/
#elif defined(SEM_TCP) && 0 /* disabled until the correct code is written */

/***********************/
#else
		wvalue = -1;
		rvalue = -1;
#endif

/******************************************************************/
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
	/************************************/
	/* calculate some interesting facts */
	/* NOTE: can only do so if value wvalues and rvalues are known */
	lock_held = (wvalue < 1);
	write_held = lock_held && ((plock->readcount < 1) || (rvalue < 1));
	read_held = lock_held && !write_held;
	if (read_held)
		read_procs_holding = plock->readcount;
	else	read_procs_holding = 0;

	procs_waiting = -wvalue;
	read_procs_waiting = -(rvalue - 1);
#  if defined(SEM_SYSVIPC)
	procs_waiting += wncnt;
	read_procs_waiting += rncnt;
#  endif
	write_procs_waiting = procs_waiting - read_procs_waiting + !lock_held;
	/* Test: */ write_procs_waiting = -(wvalue - 1) - (read_procs_waiting > 0) - lock_held;	/* NOTE: the first waiting read also gets counted in this value */
	#if defined(SEM_SYSVIPC)
		write_procs_waiting += wncnt;
	#endif
	/* Test: */ procs_waiting = read_procs_waiting + write_procs_waiting;
#endif


	/**********************************/
	/* put the status into the string */
/******************************************************************/
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
	sprintf(string,
		"%s (%s%dr%s %s%dw%s)",
		(lock_held ?
			(read_held ? RED_TEXT "%dread" NORM_TEXT :
				     RED_TEXT "write" NORM_TEXT) :
			"none "),
		(read_procs_waiting > 0 ? RED_TEXT : ""),
		read_procs_waiting,
		(read_procs_waiting > 0 ? NORM_TEXT : ""),
		(write_procs_waiting > 0 ? RED_TEXT : ""),
		write_procs_waiting,
		(write_procs_waiting > 0 ? NORM_TEXT : ""));
	if (read_held)
		sprintf(string, string, read_procs_holding);
#else /* do something simple if mutex values unknown */
	sprintf(string,
		"unknown lock status");
#endif
if (write_procs_waiting < 0) {
	vrFprintLock(stdout, lock, verbose);
}

	return string;
}


/*****************************************************************/
void vrFprintLock(FILE *file, vrLock lock, vrPrintStyle style)
{
	vrPrivateLock	*plock = (vrPrivateLock *)lock;
	int		wvalue;
	int		rvalue;
#if defined(SEM_SYSVIPC)
	int		wpid, rpid;
	int		wncnt, rncnt;
	int		wzcnt, rzcnt;
#endif
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
	int		lock_held, read_held, write_held;
	int		read_procs_holding;
	int		procs_waiting;
	int		read_procs_waiting;
	int		write_procs_waiting;
#endif

	/* TODO: print different things for different styles */
	vrTraceOpt("vrFprintLock", "beginning");

	/* if null pointer given, print an empty shell and return */
	if (plock == NULL) {
		vrFprintf(file, "Lock = { <nil> }\n");
		return;
	}

	/****************************************************************/
	/* Get the values of the semaphores based on the defined system */
/***********************/
#if defined(SEM_IRIX)
		wvalue = ustestsema(plock->wmutex);
		rvalue = ustestsema(plock->rmutex);

/***********************/
#elif defined(SEM_POSIX)
		sem_getvalue(plock->wmutex, &wvalue);
		sem_getvalue(plock->rmutex, &rvalue);

/***********************/
#elif defined(SEM_SYSVIPC)
		wvalue = semctl(plock->wmutex, 0, GETVAL);
		wpid   = semctl(plock->wmutex, 0, GETPID);
		wncnt  = semctl(plock->wmutex, 0, GETNCNT);
		wzcnt  = semctl(plock->wmutex, 0, GETZCNT);
		rvalue = semctl(plock->rmutex, 0, GETVAL);
		rpid   = semctl(plock->rmutex, 0, GETPID);
		rncnt  = semctl(plock->rmutex, 0, GETNCNT);
		rzcnt  = semctl(plock->rmutex, 0, GETZCNT);
/***********************/
#elif defined(SEM_WIN32) && 0 /* disabled until the correct code is written */

/***********************/
#elif defined(SEM_TCP) && 0 /* disabled until the correct code is written */

/***********************/
#else
		wvalue = -1;
		rvalue = -1;
#endif

/******************************************************************/
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
	/************************************/
	/* calculate some interesting facts */
	/* NOTE: can only do so if value wvalues and rvalues are known */
	lock_held = (wvalue < 1);
	write_held = lock_held && ((plock->readcount < 1) || (rvalue < 1));
	read_held = lock_held && !write_held;
	if (read_held)
		read_procs_holding = plock->readcount;
	else	read_procs_holding = 0;

	procs_waiting = -wvalue;
	read_procs_waiting = -(rvalue - 1);
#  if defined(SEM_SYSVIPC)
	procs_waiting += wncnt;
	read_procs_waiting += rncnt;
#  endif
	write_procs_waiting = procs_waiting - read_procs_waiting + !lock_held;
	/* Test: */ write_procs_waiting = -(wvalue - 1) - (read_procs_waiting > 0) - lock_held;	/* NOTE: the first waiting read also gets counted in this value */
	#if defined(SEM_SYSVIPC)
		write_procs_waiting += wncnt;
	#endif
	/* Test: */ procs_waiting = read_procs_waiting + write_procs_waiting;
#endif

	/********************/
	/* print the values */
	switch (style) {
	case one_line:
/******************************************************************/
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
		vrTraceOpt("vrFprintLock", "printing one_line");
		vrFprintf(file,
			"Lock at 0x%p, name = '%-12.12s', Status: %s (%s%dr%s %s%dw%s)\n",
			plock,
			(plock->name == NULL) ? "" : plock->name,
			(lock_held ?
				(read_held ? RED_TEXT "read " NORM_TEXT :
					     RED_TEXT "write" NORM_TEXT) :
				"none "),
			(read_procs_waiting > 0 ? RED_TEXT : ""),
			read_procs_waiting,
			(read_procs_waiting > 0 ? NORM_TEXT : ""),
			(write_procs_waiting > 0 ? RED_TEXT : ""),
			write_procs_waiting,
			(write_procs_waiting > 0 ? NORM_TEXT : ""));
#else /* do something simple if mutex values unknown */
		vrFprintf(file,
			"Lock at 0x%p, name = '%s'\n",
			plock,
			plock->name);
#endif
if (write_procs_waiting >= 0) {
		break;
}

	default:
	case verbose:
		vrTraceOpt("vrFprintLock", "printing verbose");
		vrFprintf(file, "{\n");
		vrFprintf(file,
			"\r\tObject_type = %s (%d)\n"
			"\tname = '%s'\n"
			"\ttrace = %d\n"
			"\treadcount = %d\n"
			"\ttotal read:  sets = %d releases = %d\n"
			"\ttotal write: sets = %d releases = %d\n",
			vrObjectTypeName(plock->object_type),
			plock->object_type,
			(plock->name == NULL) ? "" : plock->name,
			plock->trace,
			plock->readcount,
			plock->total_readsets, plock->total_readrels,
			plock->total_writesets, plock->total_writerels);

/***********************/
#if defined(SEM_IRIX)
		vrFprintf(file,
			"\r\twmutex = 0x%p (IRIX semvalue = %d)\n"
			  "\trmutex = 0x%p (IRIX semvalue = %d)\n",
			plock->wmutex, wvalue,
			plock->rmutex, rvalue);

/***********************/
#elif defined(SEM_POSIX)
		sem_getvalue(plock->wmutex, &wvalue);
		sem_getvalue(plock->rmutex, &rvalue);
		vrFprintf(file,
			"\r\twmutex = 0x%p (POSIX semvalue = %d)\n"
			  "\trmutex = 0x%p (POSIX semvalue = %d)\n",
			plock->wmutex, wvalue,
			plock->rmutex, rvalue);

/***********************/
#elif defined(SEM_SYSVIPC)
		vrFprintf(file,
			"\r\twmutex = 0x%p (SYSVIPC semvalue = %d, pid = %d, n = %d, z = %d)\n"
			  "\trmutex = 0x%p (SYSVIPC semvalue = %d, pid = %d, n = %d, z = %d)\n",
			plock->wmutex, wvalue, wpid, wncnt, wzcnt,
			plock->rmutex, rvalue, rpid, rncnt, rzcnt);

/***********************/
#elif defined(SEM_WIN32) && 0 /* disabled until the correct code is written */

/***********************/
#elif defined(SEM_TCP) && 0 /* disabled until the correct code is written */

/***********************/
#else
		vrFprintf(file,
			"\r\twmutex = 0x%p\n\trmutex = 0x%p\n",
			plock->wmutex,
			plock->rmutex);
#endif

/******************************************************************/
#if defined(SEM_IRIX) || defined(SEM_POSIX) || defined(SEM_SYSVIPC)
		if (lock_held) {
			if (read_held)
				vrFprintf(file, "\r\tStatus: held as " RED_TEXT "%dread" NORM_TEXT ", ", read_procs_holding);
			else	vrFprintf(file, "\r\tStatus: held as " RED_TEXT "write" NORM_TEXT ", ");
			if (read_procs_waiting != 0)
				vrFprintf(file, RED_TEXT "%d" NORM_TEXT " reads waiting, ", read_procs_waiting);
			else	vrFprintf(file, "%d reads waiting, ", read_procs_waiting);
			if (write_procs_waiting != 0)
				vrFprintf(file, RED_TEXT "%d" NORM_TEXT " writes waiting.\n", write_procs_waiting);
			else	vrFprintf(file, "%d writes waiting\n", write_procs_waiting);
		} else {
			vrFprintf(file, "\r\tStatus: no lock held\n");
		}
#endif
		vrFprintf(file, "}\n");

		break;
	}
}


/*****************************************************************/
/* Print not just the lock given, but all the subsequent locks in the linked list */
/* NOTE: we could do some tail recursion here, but to be a little more efficient, */
/*   we'll do the loop right here.                                                */
void vrFprintLockList(FILE *file, vrLock lock, vrPrintStyle style)
{
	vrPrivateLock	*plock = (vrPrivateLock *)lock;
	vrPrivateLock	*next;

	for (next = plock; next != NULL; next = next->next) {
		vrFprintLock(file, next, one_line);
	}
}




		/*********************************************/
		/*** Code Barriers structure and functions ***/
		/*********************************************/

/************************************************************************/
/*                BARRIERS                                              */
/*                Added by Ed, 11/20/1998                               */
/*                Modified by Bill, 02/28/2003                          */
/************************************************************************/

#define PRINT_FIRST_N_SYNCS	-1

#undef vrTraceOpt
#ifdef VRTRACE_BARRIER
#  define vrTraceOpt		vrTrace
#else
#  define vrTraceOpt(a,b)	;
#endif

/*****************************************************************/
/*
 * Barriers are based on the locks implemented above.  In order
 * to avoid busywaiting, a barrier is created with a dummy lock
 * (barrier_lock) to control synchronization.  This lock is initialized
 * to be writelocked.  See the comment for vrBarrierSync for more
 * details.
 */
vrBarrier *vrBarrierCreate(vrContextInfo *context, char *name, int num_clients)
{
static	char		string[128];
	vrBarrier	*b = (vrBarrier *)vrShmemAlloc0(sizeof(vrBarrier));

#if defined(SEM_SYSVIPC)
	vrMsgPrintf("FreeVR: " RED_TEXT "Reminder, barriers based on SYS-V IPC semaphores are not fully functional.\n" NORM_TEXT);
	context->startup_error |= VRSTARTUP_NOBARRSEM;
	vrSleep(1000000);
#endif
	/* set the initial barrier values */
	b->object_type = VROBJECT_BARRIER;
	b->context = context;
	b->num_clients = num_clients;
	b->num_waiting = 0;
	b->synchronizations = 0;
	b->wtime = -1.0;
	b->next = NULL;
	snprintf(string, sizeof(string), "barrier 0x%p data lock", b);
	b->lock = vrLockCreateName(context, string);
	snprintf(string, sizeof(string), "barrier 0x%p barrier lock", b);
	b->barrier_lock = vrLockCreateName(context, string);
	if (name && *name)
		b->name = vrShmemStrDup(name);
	else	b->name = "";

#if 0 /* set to 1 to trace the barrier's locks */
	vrTraceLock(b->lock, 1);
	vrTraceLock(b->barrier_lock, 1);
#endif

	/* This sets up the barrier -- all associated barriers will wait for this lock */
	vrSetWriteLock(b->barrier_lock);

	/* add this barrier to a linked list in the context */
	if (context->head_barrier == NULL) {
		/* start the list */
		context->head_barrier = b;
		context->tail_barrier = b;
	} else {
		/* append to the list */
		/* NOTE: no need to write-lock this because the call to vrBarrierCreate() */
		/*   is already inside a write lock on "context->barrier_lock"            */
		context->tail_barrier->next = b;
		context->tail_barrier = b;
	}

	vrDbgPrintfN(BARRIER_DBGLVL, "Created Barrier 0x%p, with %d clients\n", b, b->num_clients);

	return b;
}


/************************************************************/
/* Increase the number of clients associated with a barrier */
void vrBarrierIncrement(vrBarrier *b, int increment)
{
	vrSetWriteLock(b->lock);	/* lock on the data in the barrier's own struct */
	b->num_clients += increment;
	vrReleaseWriteLock(b->lock);

	vrDbgPrintfN(BARRIER_DBGLVL, "Barrier 0x%p incremented to %d clients\n", b, b->num_clients);
}


/************************************************************/
/* Decrease the number of clients associated with a barrier */
void vrBarrierDecrement(vrBarrier *b, int decrement)
{
	vrSetWriteLock(b->lock);	/* lock on the data in the barrier's own struct */
	b->num_clients += decrement;
	vrReleaseWriteLock(b->lock);

	vrDbgPrintfN(BARRIER_DBGLVL, "Barrier 0x%p decremented to %d clients\n", b, b->num_clients);
}


/*****************************************************************/
/* NOTE: It isn't clear just how appropriate this really is.  */
/*   Freeing a barrier could be a really bad idea.            */
void vrBarrierFree(vrBarrier *b)
{
	vrDbgPrintfN(BARRIER_DBGLVL, "Freeing barrier 0x%p with %d clients, %d waiting\n",
		b, b->num_clients, b->num_waiting);

	vrSetWriteLock(b->lock);	/* lock on the data in the barrier's own struct */
	vrLockFree(b->barrier_lock);
	vrLockFree(b->lock);
	vrShmemFree(b);
}


/**************************************/
/* Verify that "barrier" is a barrier */
int vrBarrierCheck(vrBarrier *barrier)
{
	if (barrier == NULL)
		return 0;

	return (barrier->object_type == VROBJECT_BARRIER);
}


/*****************************************************************/
/*
 * In normal operation, the dummy lock (barrier_lock) in a barrier
 *	is writelocked.
 * Clients wanting to synch will do one of two things:
 *	(a) all but the last client will try to readlock the barrier_lock,
 *		resulting in being suspended until that lock is unlocked
 *	(b) the last client will unwritelock the barrier_lock, resulting in
 *		all the processes from (a) being freed.
 *
 * After synching, the current (wall) time is determined and
 * placed in the barrier structure (and then returned) so all
 * processes can set the frame_wtime field in their info structure.
 * If the called with a NULL barrier, then the only thing done is
 * do determine and return the time.
 *
 * The context argument is passed for two reasons:
 *      - so each barrier has the ability to access the system context information
 *      - so each barrier will be added to the linked list of barriers for this system
 */
int vrBarrierSync(vrBarrier *b)
{
	int	mynum;		/* the order in which this process hit the barrier */

	/* if the barrier doesn't exist, then just return 0. */
	if (b != NULL) {
#ifdef VRTRACE_BARRIER
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s 0x%p\n", __FILE__, __LINE__, "vrBarrierSync", "syncing on barrier", b);
#endif
		/* NOTE: the following really should be in a read-lock, but */
		/*   since it's just debugging stuff, we'll let it slide.   */
		if (b->synchronizations < PRINT_FIRST_N_SYNCS)
			vrDbgPrintfN(BARRIER_DBGLVL, "Synchronizing on barrier 0x%p with %d clients, %d waiting + one more\n",
				b, b->num_clients, b->num_waiting);
		else if (b->synchronizations == PRINT_FIRST_N_SYNCS) {
			vrTraceLock(b->lock, 0);
			vrTraceLock(b->barrier_lock, 0);
		}

		vrSetWriteLock(b->lock);	/* lock on the data in the barrier's own struct */
		b->num_waiting++;
		mynum = b->num_waiting;
#ifdef VRTRACE_BARRIER
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s, num_waiting = %d, num_clients = %d\n", __FILE__, __LINE__, "vrBarrierSync", "checking", b->num_waiting, b->num_clients);
#endif

		if (b->num_waiting >= b->num_clients) {
			/* This is the last process to hit the barrier, so we're synced */

			if (b->synchronizations < PRINT_FIRST_N_SYNCS)
				vrDbgPrintfN(BARRIER_DBGLVL, "Barrier 0x%p has all clients, so releasing them.\n", b);

#define	BARRIER_DECREMENT	/* 6/3/03: either should work, and I think "undef" is preferred, but not sure */
#ifndef BARRIER_DECREMENT /* 3/5/3 test */
			b->num_waiting = 0;
#else
			b->num_waiting--;
#endif
			b->wtime = vrCurrentWallTime();
			b->synchronizations++;
			vrReleaseWriteLock(b->lock);

#ifdef VRTRACE_BARRIER
			vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s\n", __FILE__, __LINE__, "vrBarrierSync", "freeing");
#endif
			vrReleaseWriteLock(b->barrier_lock);	/* this frees all the waiting processes */
			vrSetWriteLock(b->barrier_lock);	/* this sets the barrier for the next pass */
		} else {
			/* We still need to wait for some other processes to check in, so wait */
			vrReleaseWriteLock(b->lock);

#ifdef VRTRACE_BARRIER
			vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s\n", __FILE__, __LINE__, "vrBarrierSync", "waiting, using read-lock on barrier lock");
#endif
			vrSetReadLock(b->barrier_lock);		/* this is where the waiting is done */
#ifdef VRTRACE_BARRIER
			vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s\n", __FILE__, __LINE__, "vrBarrierSync", "done waiting");
#endif
#ifdef BARRIER_DECREMENT
			vrSetWriteLock(b->lock);
			b->num_waiting--;
			vrReleaseWriteLock(b->lock);
#endif
			vrReleaseReadLock(b->barrier_lock);	/* we really just wanted the blocking side-effect of the lock, so release it immediately */
		}

#if 1
		/* TODO: 2/28/03 -- I don't know what these two lines are all about */
		vrSetReadLock(b->lock);
		vrReleaseReadLock(b->lock);
#endif
	} else {
		vrDbgPrintfN(BARRIER_DBGLVL+100, "NULL Barrier 0x%p so just returning 0.\n", b);
		mynum = 0;
	}

#ifdef VRTRACE_BARRIER
	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s\n", __FILE__, __LINE__, "vrBarrierSync", "beginning");
#endif
	return mynum;
}


/*****************************************************************/
/* This function returns true when all the other processes are */
/*   already waiting at this barrier, and thus the calling     */
/*   process will be the last to sync -- but might first want  */
/*   to do a little extra work just before sychronization.     */
int vrBarrierLastToSync(vrBarrier *barrier)
{
	int	last;

	if (barrier == NULL)
		return -1;

	vrSetReadLock(barrier->lock);

	last = barrier->num_waiting + 1 == barrier->num_clients;

	vrReleaseReadLock(barrier->lock);

	if (barrier->synchronizations < PRINT_FIRST_N_SYNCS)
		vrDbgPrintfN(BARRIER_DBGLVL, "Barrier 0x%p returning last-to-sync %d.\n", barrier, last);
	return last;
}


/*************************************************************/
/* This function returns true when no other processes are    */
/*   currently waiting at this barrier, and thus the calling */
/*   process will be the first to sync -- perhaps it might   */
/*   first want to do a little extra work just before        */
/*   sychronization.                                         */
int vrBarrierFirstToSync(vrBarrier *barrier)
{
	int	first;

	if (barrier == NULL)
		return -1;

	vrSetReadLock(barrier->lock);

	first = (barrier->num_waiting == 0);

	vrReleaseReadLock(barrier->lock);

	if (barrier->synchronizations < PRINT_FIRST_N_SYNCS)
		vrDbgPrintfN(BARRIER_DBGLVL, "Barrier 0x%p returning first-to-sync %d.\n", barrier, first);
	return first;
}


/*****************************************************************/
/* NOTE: this function is primarily for debugging purposes, to allow */
/*   waiting processes to continue, and then see what happens next.  */
void vrBarrierReleaseOnce(vrBarrier *barrier)
{
	/* make sure we're given a good barrier before doing anything */
	if (barrier == NULL) {
		vrPrintf("vrBarrierReleaseOnce(): NULL address given.\n");
		return;
	}

	if (barrier->object_type != VROBJECT_BARRIER) {
		vrPrintf("vrBarrierReleaseOnce(): address given is type %d, not a barrier.\n", barrier->object_type);
		return;
	}

	/* now release the barrier */
	vrSetWriteLock(barrier->lock);	/* lock on the data in the barrier's own struct */

	barrier->num_waiting = 0;
	/* NOTE: we won't increment "synchronizations" since we didn't really sync this time */
	vrReleaseWriteLock(barrier->lock);

	vrReleaseWriteLock(barrier->barrier_lock);	/* this frees all the waiting processes */
vrSetReadLock(barrier->lock);
vrReleaseReadLock(barrier->lock);
	vrSetWriteLock(barrier->barrier_lock);	/* this sets the barrier for the next pass */

	return;
}


/*****************************************************************/
void vrBarrierReleaseAll(vrContextInfo *context)
{
	vrBarrier	*barrier;

	vrDbgPrintfN(BARRIER_DBGLVL, "Releasing all barriers.\n");

	for (barrier = context->head_barrier; barrier != NULL; barrier = barrier->next) {
		vrSetWriteLock(barrier->lock);
		barrier->num_clients = 0;			/* with 0 clients, this barrier will never again bar */
		vrReleaseWriteLock(barrier->barrier_lock);	/* this frees all the waiting processes */
		/* NOTE: we also don't set the write-lock on barrier_lock again */
		/*   to prevent this barrier from doing any barriering.         */

		vrReleaseWriteLock(barrier->lock);
		/* NOTE: we can't Free the memory for this barrier inside the loop */
		/*   because we need the next pointer.                             */
	}
	vrDbgPrintfN(BARRIER_DBGLVL, "All barriers released.\n");
}


/*****************************************************************/
/* NOTE: we currently ignore the barrier's own lock -- we may want */
/*   to set the lock for reading, although considering this is     */
/*   just a debugging aid, it probably isn't necessary.            */
void vrFprintBarrier(FILE *file, vrBarrier *barrier, vrPrintStyle style)
{
	/* TODO: print different things for different styles */

	/* if null pointer given, print an empty shell and return */
	if (barrier == NULL) {
		vrFprintf(file, "Barrier = { <nil> }\n");
		return;
	}

	switch (style) {
	case one_line:
		vrFprintf(file,
			"Barrier at 0x%p, name = '%-12.12s', %d clients, %d syncs, Status: %d waiting, first? %d, last? %d\n",
			barrier,
			barrier->name,
			barrier->num_clients,
			barrier->synchronizations,
			barrier->num_waiting,
			(barrier->num_waiting == 0),
			barrier->num_waiting + 1 == barrier->num_clients);

		break;

	default:
	case verbose:
		vrFprintf(file, "{\n");
		vrFprintf(file,
			"\r\tObject_type = %s (%d)\n"
			"\tname = '%s'\n"
			"\tnext = 0x%p\n",
			vrObjectTypeName(barrier->object_type),
			barrier->object_type,
			barrier->name,
			barrier->next);
		vrFprintf(file,
			"\r\tnum_clients = %d\n\tnum_waiting = %d\n\tsynchronizations = %d\n",
			barrier->num_clients,
			barrier->num_waiting,
			barrier->synchronizations);
		vrFprintf(file,
			"\r\twtime = %.3lf\n\tlock = 0x%p\n\tbarrier_lock = 0x%p\n",
			(double)barrier->wtime,
			barrier->lock,
			barrier->barrier_lock);
		vrFprintf(file, "}\n");

		break;
	}
}


/*****************************************************************/
/* Print not just the barrier given, but all the subsequent barriers in the linked list */
/* NOTE: we could do some tail recursion here, but to be a little more efficient, */
/*   we'll do the loop right here.                                                */
void vrFprintBarrierList(FILE *file, vrBarrier *barrier, vrPrintStyle style)
{
	vrBarrier	*next;

	for (next = barrier; next != NULL; next = next->next) {
		vrFprintBarrier(file, next, one_line);
	}
}


#if 0 /* { */ /* TODO: delete this section -- I don't think we need the vrInitBarriers() or vrSysBarrierSync() functions. */

/*****************************************************************/
/*
 *
 * The system maintains one barrier for each process group;  this
 * is an API for synching with them easily.  Asking to synch with
 * process group 0 is special;  see the notes above.
 *
 */
static int sys_barrier_count;
static vrBarrier **sys_barriers;

void vrInitBarriers(int num_barriers, int *num_clients)
{
	int i;

#if 0
	sys_barrier_count = num_barriers;
	sys_barriers = vrShmemAlloc0(num_barriers * sizeof(vrBarrier*));
	for (i = 0; i < num_barriers; i++)
		sys_barriers[i] = vrCreateBarrier(num_clients[i]);
#endif
}


/*****************************************************************/
void vrSysBarrierSync(int which)
{
	vrBarrier	*b = NULL;

	if (which > 0 && which <= sys_barrier_count)
		b = sys_barriers[which-1];

	vrBarrierSync(b);
}
#endif /* } */


#ifdef SEM_WIN32 /* { */
/*****************************************************************/
static void _WaitForSemaphore(HANDLE handle)
{
	if (WaitForSingleObject(handle, INFINITE) == WAIT_FAILED) {
		LPVOID	lpMsgBuf;

		if (!FormatMessage(
					FORMAT_MESSAGE_ALLOCATE_BUFFER |
					FORMAT_MESSAGE_FROM_SYSTEM |
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,
					GetLastError(),
					MAKELANGID(LANG_NEUTRAL,
						SUBLANG_DEFAULT),
					(LPTSTR) &lpMsgBuf,
					0,
					NULL)) {
			vrErr("FormatMessage() failed");
			vrErrPrintf("Waiting failed for lock (%p)\n", (void *)handle);
		} else {
			vrErrPrintf("Waiting failed for lock (%p)\n"
					"GetLastError() = %s\n",
					(void *)handle,
					(LPCTSTR)lpMsgBuf);
			LocalFree(lpMsgBuf);
		}
	}
}
#endif /* } */



#if !defined(SEM_WIN32) && !defined(SEM_TCP) /* { */
/* The following functions break the CYGWIN build */

/* TODO: these functions should probably be moved to another file */

/*****************************************************************/
/*===============================================================*/
/*****************************************************************/
/* The following routines are for key-based shared memory access.*/


#include <sys/types.h>
#if !(defined(SHM_SYSVIPC) || defined(SEM_SYSVIPC)) /* ie. already included */
#  include <sys/ipc.h>	/* IRIX 6.x includes this in sys/shm.h, but not 5.3 */
#  include <sys/shm.h>
#endif


#define PERMS 0666


/*************************************************************************
 int vrShmemGetKey()
	Gets a shared memory segment -- creates if necessary.
**************************************************************************/
int vrShmemGetKey(key_t shmem_key, int memsize, void **memptr, int create)
{
	int	shmid;
unsigned int	sh_flags;
	void	*memory;

	if (create)
		sh_flags = PERMS | IPC_CREAT;
	else 	sh_flags = (unsigned int)NULL;

	shmid = shmget(shmem_key, memsize, sh_flags);
	if (shmid < 0) {
		/* TODO: need to check for cause of failure -- e.g. permissions, ... */
		if (create) {
			perror("shmget");
			fprintf(stderr,"can't get shared memory: key %d\n", shmem_key);
			exit(-1);	/* TODO: this seems a bit harsh -- let the program decide */
		} else {
			*memptr = NULL;
			return shmid;
		}
	}

	memory = (void *)shmat(shmid, (char *)0, 0);
	if (memory == (void *)-1) {
		perror("shmat");
		fprintf(stderr,"can't attach shared memory: key %d\n", shmem_key);
		exit(-1);	/* TODO: this seems a bit harsh -- let the program decide */
	}
	*memptr = memory;

	/* TODO: it might be wise to print a warning message if the requested */
	/*   memory size doesn't match the memory size of the given segment.  */
	/* TODO: for debugging it might be nice to have a vrShmemPrintId()  */
	/*   function that prints the data associated with a given segment. */

	return shmid;
}


/*************************************************************************/
int vrShmemIdGetSize(int shmid)
{
	struct shmid_ds		shm_data;	/* structure with info about shared memory segment */
	int			ctl;		/* for interfacing with shared memory controls     */

	/* get the statistics on the shared memory segment */
	errno = 0;
	ctl = shmctl(shmid, IPC_STAT, &shm_data);

	if (ctl < 0) {
		/* print warnings/error messages for possible errors */
		switch (errno) {
		case EACCES:	/* permission denied */
			vrErrPrintf("vrShmemIdGetSize(): " RED_TEXT "ERROR: permission denied to shared memory id %d\n" NORM_TEXT, shmid);
			break;

		case EINVAL:	/* shmid is not valid */
			vrErrPrintf("vrShmemIdGetSize(): " RED_TEXT "ERROR: invalid shared memory id %d\n" NORM_TEXT, shmid);
			break;

#if defined(EOVERFLOW)
		case EOVERFLOW:	/* status cannot be stored */
			vrErrPrintf("vrShmemIdGetSize(): " RED_TEXT "ERROR: overflow error -- status cannot be acquired for shmid %d\n" NORM_TEXT, shmid);
			break;
#endif

		case EFAULT:	/* illegal data address (ie. of shm_data, which shouldn't happen) */
			vrErrPrintf("vrShmemIdGetSize(): " RED_TEXT "ERROR: illegal data address while reading shmid %d\n" NORM_TEXT, shmid);
			break;
		}

		return -1;
	}

	return shm_data.shm_segsz;
}


/*************************************************************************
 void vrShmemDetach()
	Get the shared memory id and remove it.
**************************************************************************/
void vrShmemDetach(void *memory_address)
{
	if (shmdt(memory_address) < 0)
		perror("Detaching shared memory");

	/* TODO: this function.  probably "shmdt(memory_address);" */
}


/*************************************************************************
 void vrShmemRemoveKey()
	Get the shared memory id and remove it.
**************************************************************************/
void vrShmemRemoveKey(key_t shmem_key)
{
	int	shmid;

	shmid = shmget(shmem_key, (int)NULL, (int)NULL);
	if (shmid > -1) {
		/* only linux needs the third argument to shmctl */
		if (shmctl(shmid, IPC_RMID, (struct shmid_ds *)NULL) < 0)
			perror("Removing shmid");
	}
}


/*************************************************************************
 void vrShmemRemoveId()
	Remove shared memory by id.
**************************************************************************/
void vrShmemRemoveId(int shmid)
{
	if (shmid > -1) {
		if (shmctl(shmid, IPC_RMID, (struct shmid_ds *)NULL) < 0)
			perror("Removing shmid");
	}
}

#endif /* } */

