/* ======================================================================
 * 
 *  CCCCC          vr_socket.c
 * CC   CC         Author(s): Bill Sherman, John Stone
 * CC              Created: October 22, 2001  (from code originally written in 1987)
 * CC   CC         Last Modified: April 25, 2002
 *  CCCCC
 * 
 * Code file for FreeVR socket interfacing.
 * 
 * Copyright 2002, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/****************************************************************************/
/* Socket API by Bill Sherman                                               */
/*                                                                          */
/* This version is for TCP style sockets only.                              */
/*                                                                          */
/* ROUTINES:                                                                */
/*   int vrSocketCreateListen(port, block)                                  */
/*   void vrSocketClose(sock_fd)                                            */
/*   int vrSocketAnswer(sock_fd)                                            */
/*   int vrSocketCall(hostname, port)                                       */
/*   int vrSocketReadMsg(sock_fd, msg, msglen)                              */
/*   int vrSocketReadMsgClean(sock_fd, msg, msglen)                         */
/*   int vrSocketSendMsg(sock_fd, msg)                                      */
/*                                                                          */
/* HISTORY:                                                                 */
/*  1987/8: original version(s) isockets.c and usockets.c written by Bill   */
/*	Sherman as alternative (IP vs. Unix) socket routines for a series   */
/*	of programs for card playing over the Internet (way back in 1987).  */
/*	The Cards programs were written specifically to learn how to        */
/*	write sockets and curses programs.                                  */
/*...                                                                       */
/*  Feb 1998: renamed isockets.c to socket_api.c and refined the interface. */
/*                                                                          */
/*  10/22/2001: incorporated into FreeVR library as vr_socket.c, with       */
/*	functions renamed to fit the FreeVR naming scheme.                  */
/*                                                                          */
/* TODO:                                                                    */
/*   - Describe each function and its arguments in comments.                */
/*                                                                          */
/****************************************************************************/
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/file.h>		/* added by Matt Schechtman for Apple? */
#include <netinet/in.h>
#include <fcntl.h>		/* added by Matt Schechtman for Apple? */
#include <netdb.h>
#include <string.h>		/* added by Matt Schechtman for strlen */
#include <strings.h>		/* added by Matt Schechtman for bzero */
#include <unistd.h>		/* added by Matt Schechtman for read/write */
#include <stdio.h>

#include "vr_socket.h"		/* really this is just to make sure function args match */


/*************************************************************************/
/* vrSocketCreateListen(): ...    */
/* port is used to assign a port.  If port==0, then an unassigned socket */
/*   is chosen, and the result is passed back via the port variable.     */
/*************************************************************************/
int vrSocketCreateListen(port, block)
	int	*port;		/* port is returned via this CBR variable */
	int	block;		/* flag to indicate whether to block (1) or not (0) */
{
struct	sockaddr_in	server_addr;		/* TODO: should this be "sockaddr" type? */
	int		sock_fd,
			fd_flags, fd_flags_block, fd_flags_noblock,
			length;

	/* Create socket */
	sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sock_fd < 0) {
		perror("opening stream socket");
		return -1;
	}

	/* set socket port to blocking or non-blocking as indicated */
	fd_flags = fcntl(sock_fd, F_GETFL, 0);
	fd_flags_block = fd_flags & ~FNDELAY;  /* Turn blocking on */
	fd_flags_noblock = fd_flags | FNDELAY; /* Turn blocking off */
	fcntl(sock_fd, F_SETFL, block ? fd_flags_block : fd_flags_noblock);

	/* Name socket using inet port scheme */
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	server_addr.sin_port = htons(*port);
	if (bind(sock_fd, (struct sockaddr *)&server_addr, sizeof(server_addr))) { /* TODO: why is this "sockaddr", when the type is "sockaddr_in" ?? */
		perror("binding stream socket");
		close(sock_fd);
		return -1;
	}

	length = sizeof(server_addr);
	if (getsockname(sock_fd, (struct sockaddr *)&server_addr, &length)) { /* TODO: why is this "sockaddr", when the type is "sockaddr_in" ?? */
		perror("getting socket name");
		return -1;
	}
	*port = ntohs(server_addr.sin_port);
	listen(sock_fd, 5);
	return sock_fd;
}


/******************************************************/
/* vrSocketClose(): ... */
/******************************************************/
void vrSocketClose(sock_fd)
	int	sock_fd;
{
	close(sock_fd);
	shutdown(sock_fd, 2);
}


/******************************************************/
/* vrSocketAnswer(): ... */
/******************************************************/
/* if vrSocketAnswer() returns -1, then no connection is made -- which  */
/*    is fine for non-blocking sockets.                                */
int vrSocketAnswer(sock_fd)
	int	sock_fd;
{
	int	asock;

	asock = accept(sock_fd, 0, 0);

	return asock;
}


/******************************************************/
/* vrSocketCall(): ... */
/******************************************************/
/* this routine is used by the client end of a connection */
int vrSocketCall(hostname, port /* , block */)
	char	*hostname;
	int	port;
#if 0
	int	block; /* not sure if this will be useful */
#endif
{
struct	sockaddr_in	client_addr;		/* TODO: should this be "sockaddr" type? */
struct	hostent		*hp,
			*gethostbyname();
	int		sock_fd;

	/* Create socket */
	sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sock_fd < 0) {
		perror("opening stream socket");
		return -1;
	}

	/* Name socket using file system name */
	hp = gethostbyname(hostname);
	if (hp == 0) {
		fprintf(stderr, "%s: unknown host\n", hostname);
		return -1;
	}
	bcopy(hp->h_addr, &client_addr.sin_addr, hp->h_length);

	client_addr.sin_family = AF_INET;
	client_addr.sin_port = htons(port);

	if (connect(sock_fd, (struct sockaddr *)&client_addr, sizeof(struct sockaddr_in)) < 0) { /* TODO: why is this "sockaddr", when the type is "sockaddr_in" ?? */
		close(sock_fd);
		perror("connecting stream socket");
		return -1;
	}
	return sock_fd;
}


/******************************************************/
/* vrSocketReadMsg(): ... */
/******************************************************/
int vrSocketReadMsg(sock_fd, msg, msglen)
	int	sock_fd;
	char	*msg;
	int	msglen;
{
	int	rval;

	bzero(msg, msglen);

	rval = read(sock_fd, msg, msglen);

	return rval;
}


/******************************************************/
/* vrSocketReadMsgClean(): ... */
/* same as vrSocketReadMsg, but strips tail '\n' */
/******************************************************/
int vrSocketReadMsgClean(sock_fd, msg, msglen)
	int	sock_fd;
	char	*msg;
	int	msglen;
{
	int	rval;


	bzero(msg, msglen);
	rval = read(sock_fd, msg, msglen);

	/* clean off trailing '\n' */
	if (msg[rval-1] == '\012')
		msg[rval-1] = '\0';
	if (msg[rval-2] == '\r')
		msg[rval-2] = '\0';

	return rval;
}


/******************************************************/
/* vrSocketSendMsg(): ... */
/******************************************************/
int vrSocketSendMsg(sock_fd, msg)
	int	sock_fd;
	char	*msg;
{
	int	rval;

#if 0 /* this is basically a nop (or should be) but for some reason causes a seg fault on Linux */
	/* TODO: figure out why this seg faults on Linux */
	msg[strlen(msg)] = '\0';
#endif

	rval = write(sock_fd, msg, strlen(msg)+1);
	if (rval < 0)
		perror("vrSocketSendMsg");

	return rval;
}

