/* ======================================================================
 * 
 * HH   HH         vr_visren.h
 * HH   HH         Author(s): Ed Peters, Bill Sherman
 * HHHHHHH         Created: June 4, 1998
 * HH   HH         Last Modified: June 5, 2003
 * HH   HH
 * 
 * Header file for FreeVR visual rendering information.
 * 
 * Each Window onto a VE has a certain number of eyes for which to
 * render.  Each Window has its own (default) rendering function,
 * which might be overridden by a particular user of that window.
 * Each window also has projection coordinates, and a viewport mask.
 * 
 * In order to at least maintain the fiction of window-system independence,
 * the really GLX-specific code is isolated in vr_visren.glx.*
 * 
 * Copyright 2003, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#ifndef __VRVISREN_H__
#define __VRVISREN_H__

#include <sys/types.h>
#include <unistd.h>

#include "vr_callback.h"
#include "vr_procs.h"

#ifdef __cplusplus
extern "C" {
#endif

#define VR_MAXSTATS	5


/**************************************************************/
/* The four ways in which a window can be placed in the world */
typedef enum vrWindowMountType_en {
		VRWINDOW_FIXED,
		VRWINDOW_HEADMOUNT,
		VRWINDOW_HANDHELD,
		VRWINDOW_SIMULATOR
	} vrWindowMountType;


/****************************************************************************/
/* TODO: consider making these an enumerated type */
/*   NOTE: these are used by the vrSimulatorMove() function */
#define VR_SIMMOVE_NOP		 0
#define VR_SIMMOVE_AWAY		 1
#define VR_SIMMOVE_TOWARD	 2
#define	VR_SIMMOVE_CENTER	 3
#define	VR_SIMMOVE_SETHOME	 4
#define	VR_SIMMOVE_GOHOME	 5
#define	VR_SIMMOVE_POSY		 6
#define	VR_SIMMOVE_NEGY		 7
#define	VR_SIMMOVE_POSX		 8
#define	VR_SIMMOVE_NEGX		 9
#define	VR_SIMMOVE_TOGGLEMASK	10


/********************************************************************************/
/* A rectangle viewport for specifying where a window viewport should be placed */
typedef struct {
		int		origX;
		int		origY;
		int		width;
		int		height;
	} vrViewport;


/****************************************************************************/
/* A rectangle *fractional* viewport for specifying in which subregion of a */
/*   window the viewport should be placed                                   */
typedef struct {
		float		min_X;
		float		max_X;
		float		min_Y;
		float		max_Y;
	} vrViewportFract;


/*************************************************************************************/
/* (currently unused) some data that indicates what part of the viewport to mask off */
typedef struct {
		void		*data;
	} vrViewportMask;


/***************************************************************/
/* the data used to calculate the perspective rendering matrix */
typedef struct {
		union	{
			double		v[6];		/* access to frustum values as an array (vector) */
			vrFrustum	n;		/* access to frustum values by name */
		} frustum;
		double			eye[3];
		vrMatrix		mat;		/* NOTE: if filled in, this will be redundant */
							/*   data to the frust/eye combination, and   */
							/*   problably won't be used much.            */
	} vrPerspData;
#define	USE_FRUSTUMEYE	/* should be defined -- undef to have FreeVR calc 4x4 persp matrix */


/***************************************************************************/
/* Fields marked as "CONFIG" are filled in direction from the config       */
/*   file.  Other fields are determined in combination with other factors. */
typedef struct vrWindowInfo_st {

		/*************************/
		/* Generic Object fields */
		vrObject	object_type;	/* The type of this object (ie. VROBJECT_WINDOW) */
		int		id;		/* numeric id for this window. */
		char		*name;		/* CONFIG: The name assigned to this window. */
		int		malleable;	/* CONFIG: Whether the data of this window can be modified */
	struct	vrWindowInfo_st	*next;		/* the next window in a linked list */
		vrContextInfo	*context;	/* pointer to the vrContext memory structure */
		int		line_created;	/* line number of where this object was created in config */
		int		line_lastmod;	/* line number of last time this object modified in config */
		char		file_created[512];/* file (or other) where this object was created*/
		char		file_lastmod[512];/* file (or other) where this object was created*/

		/******************************/
		/* Fields specific to Windows */
		char		*version;	/* The version of this window */
		vrProcessInfo	*proc;		/* Pointer to the process handling this window */
		void		*aux_data;	/* auxillary data, specific to the type of window */
		vrSystemSettings settings;	/* CONFIG: values for many inheritable fields */
		char		*graphics;	/* CONFIG: type of graphics renderer (eg. "GLX") */
		char		*args;		/* CONFIG: arguments for how this window should look/be rendered */

		vrWindowMountType mount;	/* CONFIG: The mount type of this window (eg. fixed, head-based) */
		int		num;		/* the number of this window in running system (1-based) */
		int		dualeye_buffer;	/* Boolean that indicates whether window has LEFT & RIGHT eye buffers */
		vrVisrenModeType visrenmode;	/* The type of visual rendering mode selected */
		int		num_eyes;	/* CONFIG: The number of eyes for this window */
		vrEyelistInfo	*eyelist;	/* Pointer to the eyelist given by settings.eyelist_name */
		vrEyeInfo	**eyes;		/* Array of pointers to the eyes rendered by this window */
		struct vrUser_st *sim_user;	/* Pointer to the user that simulator view is tethered to */
		double		coords_ll[3];	/* CONFIG: The location of the lower-left corner of the window.   */
						/*   (relative to fixed, head, hand movement based on mount type. */
		double		coords_lr[3];	/* CONFIG: The location of the lower-right corner of the window.  */
		double		coords_ul[3];	/* CONFIG: The location of the upper-left corner of the window.   */
#if 1
		vrMatrix	*OLD_world_to_window;	/* TODO: delete this field */
#endif
		vrPoint		Lpoint;		/* lower left corner point of window */
		vrPoint		Hpoint;		/* upper right corner point of window */
		double		H_minus_L[2];	/* values calculated based on "coords" fields  */
		double		H_plus_L[2];	/* values calculated based on "coords" fields  */
		double		HminusL_recip[2];/* values calculated based on "coords" fields */
		vrMatrix	*e2w_xform;	/* A matrix that maps eye coords to window coords */
		vrMatrix	*rw2w_xform;	/* A matrix that maps real-world coordinates to window coords */
		vrMatrix	*rw2w_homexform;/* Home rw2w matrix for simulator view */
		vrViewport	geometry;	/* CONFIG: The window's geometry on the window-manager. */
		vrViewport	viewport_left;	/* CONFIG: left viewport in OpenGL parameters */
		vrViewport	viewport_right;	/* CONFIG: right viewport in OpenGL parameters */
		vrViewportFract	viewportF_left;	/* CONFIG: fractional viewport for left view for dualvp mode */
		vrViewportFract	viewportF_right;/* CONFIG: fractional viewport for right view for dualvp mode */
		vrViewportMask	viewport_mask;	/* CONFIG: TODO: A viewport mask to constrain rendering in the window. */
		int		simulator_mask;	/* Changable/CONFIG: a bitmask indicating what simulator parts to display. */

		int		fps_show;	/* Changable/CONFIG: A flag indicating to put fps on window. */
		float		fps_loc[2];	/* Changable/CONFIG: where on window to put fps info */
		float		fps_color[3];	/* Changable/CONFIG: color to display the fps data */
		int		stats_show;	/* Changable/CONFIG: A flag/mask indicating what stats to put on window. */
		char		*stats_procs;	/* CONFIG: a string of which processes to show stats */
		vrProcessStats	**stats[VR_MAXSTATS];/* array of pointers to pointers to stats information from stats_procs */

		/********************************/
		/* Callbacks for what to render */
		int		call_visreninit;/* Flag to indicate whether the current VisrenInit callback has been called for this window */
		vrCallback	*VisrenInit;	/* function to initialize the graphics in a rendering window */
		vrCallback	*VisrenFrame;	/* function to set state information for a rendering frame */
		vrCallback	*VisrenWorld;	/* function to render the virtual world scene (called per eye) */
		vrCallback	*VisrenSim;	/* function to render the simulator view */
		vrCallback	*VisrenExit;	/* function to restore any graphics state prior to exitting */

		/*******************************/
		/* Callbacks for how to render */
		/* TODO: perhaps these should be system-wide config callbacks, not per-window callbacks. */
		vrCallback	*PreOpenInit;	/* callback for graphics systems that require info before opening windows (eg. pf) */
		vrCallback	*Open;		/* function to open the rendering window */
		vrCallback	*Render;	/* function to render the scene */
		vrCallback	*RenderText;	/* function to render text at the current 3D location */
		vrCallback	*RenderNullWorld;/* function to render an empty world */
		vrCallback	*RenderSimulator;/* function to render the simulated VR hardware */
		vrCallback	*RenderTransform;/* function to make a 3D transformation */
		vrCallback	*Swap;		/* function to swap front and back graphics buffers */
		vrCallback	*Close;		/* function to close the rendering window */
		vrCallback	*PrintAux;	/* function to print the window's auxillary data */

	} vrWindowInfo;


/****************************************************************************/
/* Array indices for different time measurement values */
#define	VR_TIME_INIT	 0
#define	VR_TIME_FRAME	 1
#define	VR_TIME_RENDER	 2
#define	VR_TIME_WAIT	 3
#define	VR_TIME_SYNC	 4
#define	VR_TIME_FREEZE	 5
#define	VR_TIME_SWAP	 6
#define	VR_TIME_TRAVEL	 7

#if 0 /* deprecated by new stats stuff */
#define	VR_TIME_SHOW	 8		/* NOTE: this also indicates the size of the array */
#define	VR_TIME_CALC	 9		/* flag that indicates whether to even calculate times */
#define	VR_TIME_DEFAULT	 0x3ff		/* bits to turn everything on */

#ifdef __linux
#  define	VR_TIME_FRAMES	200
#else
#  define	VR_TIME_FRAMES	100
#endif
#endif /* end-deprecated section */



/*******************************************************/
/* type definition for data passed to each render call */
typedef struct {
		vrContextInfo	*context;	/* Pointer to overall system context     */
		vrPerspData	*persp;		/* Pointer to how to render this frame   */
		vrWindowInfo	*window;	/* Pointer to window rendering into      */
		vrEyeInfo	*eye;		/* Pointer to eye rendering to           */
		long		frame_count;	/* This process' frame_count             */
		vrTime		frame_stime;	/* Sim-Time when this render frame began */
#if 0 /* deprecated */
		int		time_frame;	/* current incoming measures element   */
		vrTime		measures[VR_TIME_FRAMES][VR_TIME_SHOW];/* array of times for current frame */
		vrProcessStats	**stats[VR_MAXSTATS];	/* an array of pointers to where the pointers to stats information may be */
#endif /* end-deprecated */
	} vrRenderInfo;


/*****************************************************************/
/* vrVisrenOptsType: structure for adding new rendering options. */
typedef struct {
		char	*option_name;
		void	(*info_func)(vrWindowInfo *);
		void	(*nullworld_func)();
		void	(*defsim_func)();
		void	(*rentext_func)(char *);
	} vrVisrenOptsType;


/*************************************************/
/*** Globally accessible function declarations ***/

void		vrVisrenMainLoop(vrProcessInfo *);

char		*vrWindowName(vrWindowMountType type);
vrWindowMountType vrWindowType(char *name);
char		*vrVisrenModeName(vrVisrenModeType mode);
#if 0 /* this should work */
vrVisrenModeType vrVisrenModeValue(char *modename);
#else
int		 vrVisrenModeValue(char *modename);
#endif
void		vrWindowClear(vrWindowInfo *object);
void		vrWindowCopy(vrWindowInfo *dest_object, vrWindowInfo *src_object);
void		vrFprintWindowInfo(FILE *, vrWindowInfo *, vrPrintStyle style);
vrWindowInfo	*vrCurrentWindow();
struct vrUser_st *vrCurrentUser();
vrEyeInfo	*vrCurrentEye();
void		vrVisrenMainLoop(vrProcessInfo *myproc_info);
void		vrGetWindowInfo(vrWindowInfo *);
void		vrRenderNullWorld(void);
void		vrRenderDefaultSimulator(vrRenderInfo *renderinfo, int mask);
void		vrRenderText(char *string);
void		vrRenderTransform(vrMatrix *mat);
void		vrCalcPerspIntermediaries(vrWindowInfo *window);
#ifdef USE_FRUSTUMEYE
vrPerspData	*vrCalcPerspFrustumEye(vrPerspData *pd, vrWindowInfo *this_window, vrMatrix *eye_rwpos);
#else
vrPerspData	*vrCalcPerspMatrix(vrPerspData *pd, vrWindowInfo *this_window, vrMatrix *eye_rwpos);
#endif
void		vrSimulatorMove(vrWindowInfo *this_window, int command, void *args);
void		vrSimulatorMoveString(vrWindowInfo *window, char *strcommand);

#ifdef __cplusplus
}
#endif

#endif
